from dataclasses import dataclass
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from pypika.terms import Field
from pypika.terms import Term
from pypika.terms import WindowFrameAnalyticFunction
from typeguard import typechecked

from tecton_core import data_types as tecton_types
from tecton_core.data_types import ArrayType
from tecton_core.data_types import DataType
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int64Type
from tecton_proto.args import feature_view__client_pb2 as feature_view__args_pb2
from tecton_proto.common import aggregation_function__client_pb2 as afpb


# Maps an aggregation proto to its respective simple string function name.
AGGREGATION_FUNCTIONS_TO_COLUMN_NAME = {
    afpb.AGGREGATION_FUNCTION_COUNT: "count",
    afpb.AGGREGATION_FUNCTION_SUM: "sum",
    afpb.AGGREGATION_FUNCTION_MEAN: "mean",
    afpb.AGGREGATION_FUNCTION_LAST: "last",
    afpb.AGGREGATION_FUNCTION_MIN: "min",
    afpb.AGGREGATION_FUNCTION_MAX: "max",
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: "var_samp",
    afpb.AGGREGATION_FUNCTION_VAR_POP: "var_pop",
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: "stddev_samp",
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: "stddev_pop",
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: "last_non_distinct_n",
    afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N: "lastn",
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: "first_non_distinct_n",
    afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N: "first_distinct_n",
    afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT: "approx_count_distinct",
    afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE: "approx_percentile",
}

# Maps a simple string aggregation function used to define feature views to its respective aggregation function proto.
AGGREGATION_FUNCTION_STR_TO_ENUM = {
    "stddev": afpb.AggregationFunction.AGGREGATION_FUNCTION_STDDEV_SAMP,
    "stddev_samp": afpb.AggregationFunction.AGGREGATION_FUNCTION_STDDEV_SAMP,
    "last": afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST,
    "count": afpb.AggregationFunction.AGGREGATION_FUNCTION_COUNT,
    "mean": afpb.AggregationFunction.AGGREGATION_FUNCTION_MEAN,
    "min": afpb.AggregationFunction.AGGREGATION_FUNCTION_MIN,
    "max": afpb.AggregationFunction.AGGREGATION_FUNCTION_MAX,
    "var_pop": afpb.AggregationFunction.AGGREGATION_FUNCTION_VAR_POP,
    "var_samp": afpb.AggregationFunction.AGGREGATION_FUNCTION_VAR_SAMP,
    "variance": afpb.AggregationFunction.AGGREGATION_FUNCTION_VAR_SAMP,  # variance is a var_samp alias.
    "stddev_pop": afpb.AggregationFunction.AGGREGATION_FUNCTION_STDDEV_POP,
    "sum": afpb.AggregationFunction.AGGREGATION_FUNCTION_SUM,
    "lastn": afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_DISTINCT_N,
    "last_non_distinct_n": afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N,
    "first_non_distinct_n": afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N,
    "first_distinct_n": afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_DISTINCT_N,
    "approx_count_distinct": afpb.AggregationFunction.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT,
    "approx_percentile": afpb.AggregationFunction.AGGREGATION_FUNCTION_APPROX_PERCENTILE,
}


def get_aggregation_enum_from_string(aggregation_function: str) -> afpb.AggregationFunction:
    aggregation_function_enum = AGGREGATION_FUNCTION_STR_TO_ENUM.get(aggregation_function, None)
    if aggregation_function_enum is None:
        msg = f"Unsupported aggregation function {aggregation_function}"
        raise ValueError(msg)
    return aggregation_function_enum


def get_aggregation_function_name(aggregation_function_enum):
    return AGGREGATION_FUNCTIONS_TO_COLUMN_NAME[aggregation_function_enum]


# Column prefixes that can't be derived from aggregation function name.
sum_of_squares_column_prefix = get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_SUM) + "_of_squares"


def get_pretty_column_prefix(materialized_column_prefix: str) -> str:
    """Formats a materialized column prefix to be more human readable.

    For example, maps 'lastn10' to 'last_distinct_10' and 'last_non_distinct_n10' to 'last_10'.

    Only used for cosmetic purposes, to prettify the materialized column prefixes in the run API.
    """
    if materialized_column_prefix.startswith("lastn"):
        n = materialized_column_prefix.lstrip("lastn")
        return f"last_distinct_{n}"
    elif materialized_column_prefix.startswith("last_non_distinct_n"):
        n = materialized_column_prefix.lstrip("last_non_distinct_n")
        return f"last_{n}"
    elif materialized_column_prefix.startswith("first_non_distinct_n"):
        n = materialized_column_prefix.lstrip("first_non_distinct_n")
        return f"first_{n}"
    elif materialized_column_prefix.startswith("first_distinct_n"):
        n = materialized_column_prefix.lstrip("first_distinct_n")
        return f"first_distinct_{n}"

    return materialized_column_prefix


def get_materialization_aggregation_column_prefixes(
    aggregation_function_enum: afpb.AggregationFunction,
    function_params: Optional[
        Union[Dict[str, feature_view__args_pb2.ParamValue], afpb.AggregationFunctionParams]
    ] = None,
    is_continuous: bool = False,
) -> List[str]:
    """Maps an aggregation function to the prefixes that should be applied to the materialized columns.

    For example, the "sum" partial aggregation requires only a single column prefix, "sum", while the "mean" partial
    aggregation requires two column prefixes, "mean" and "count".

    The order of the column prefixes is important; it must match the order of the columns computed by the partial
    aggregate methods.

    For most aggregations, the prefixes should be the same for continuous and non-continuous.
    """
    if aggregation_function_enum not in _AGGREGATION_COLUMN_PREFIX_MAP:
        msg = f"Unsupported aggregation {aggregation_function_enum}"
        raise ValueError(msg)
    prefixes = _AGGREGATION_COLUMN_PREFIX_MAP[aggregation_function_enum]

    if not is_continuous and aggregation_function_enum in (
        afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_DISTINCT_N,
        afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N,
    ):
        if isinstance(function_params, afpb.AggregationFunctionParams):
            return [prefixes[0] + str(function_params.last_n.n)]

        assert function_params is not None, "function_params must be set for last N aggregations in non-continuous mode"
        return [prefixes[0] + str(function_params["n"].int64_value)]
    elif not is_continuous and aggregation_function_enum in (
        afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N,
        afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_DISTINCT_N,
    ):
        if isinstance(function_params, afpb.AggregationFunctionParams):
            return [prefixes[0] + str(function_params.first_n.n)]

        assert (
            function_params is not None
        ), "function_params must be set for first N aggregations in non-continuous mode"
        return [prefixes[0] + str(function_params["n"].int64_value)]

    return prefixes


# Sample and Population Standard Deviation and Variance only depend on sum of squares, count, and sum. For example, to
# calculate population variance you can divide the sum of squares by the count and subtract the square of the mean.
_var_stddev_prefixes = [
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_SUM) + "_of_squares",
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT),
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_SUM),
]

_approx_count_distinct_prefixes = [
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT) + "_indices",
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT) + "_registers",
]

_approx_percentile_prefixes = [
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE) + "_processed_means",
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE) + "_processed_weights",
]

_AGGREGATION_COLUMN_PREFIX_MAP = {
    afpb.AGGREGATION_FUNCTION_SUM: [get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_SUM)],
    afpb.AGGREGATION_FUNCTION_MIN: [get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MIN)],
    afpb.AGGREGATION_FUNCTION_MAX: [get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MAX)],
    afpb.AGGREGATION_FUNCTION_COUNT: [get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT)],
    afpb.AGGREGATION_FUNCTION_LAST: [get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST)],
    afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N: [
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N)
    ],
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: [
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N)
    ],
    afpb.AGGREGATION_FUNCTION_MEAN: [
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MEAN),
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT),
    ],
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: _var_stddev_prefixes,
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: _var_stddev_prefixes,
    afpb.AGGREGATION_FUNCTION_VAR_POP: _var_stddev_prefixes,
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: _var_stddev_prefixes,
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: [
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N)
    ],
    afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N: [
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N)
    ],
    afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT: _approx_count_distinct_prefixes,
    afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE: _approx_percentile_prefixes,
}


@typechecked
def aggregation_prefix_to_tecton_type(
    prefix: str, input_type: tecton_types.DataType
) -> Optional[tecton_types.DataType]:
    prefix = prefix.lower()
    if prefix == "count":
        return tecton_types.Int64Type()
    elif prefix == "mean" or prefix == "sum_of_squares":
        return tecton_types.Float64Type()
    elif prefix.startswith(("lastn", "last_non_distinct_n", "first_non_distinct_n", "first_distinct_n")):
        return tecton_types.ArrayType(input_type)
    elif prefix.startswith(tuple(_approx_count_distinct_prefixes)):
        return tecton_types.ArrayType(tecton_types.Int64Type())
    elif prefix.startswith(tuple(_approx_percentile_prefixes)):
        return tecton_types.ArrayType(tecton_types.Float64Type())
    else:
        return None


@typechecked
def get_materialization_column_name(prefix: str, input_column_name: str) -> str:
    return prefix + "_" + input_column_name


# reference doc: https://docs.tecton.ai/docs/defining-features/feature-views/aggregation-engine/aggregation-functions
_AGGREGATION_RESULT_TYPE: Dict[int, Callable[[DataType], DataType]] = {
    afpb.AGGREGATION_FUNCTION_SUM: lambda feature_type: feature_type,
    afpb.AGGREGATION_FUNCTION_MIN: lambda feature_type: feature_type,
    afpb.AGGREGATION_FUNCTION_MAX: lambda feature_type: feature_type,
    afpb.AGGREGATION_FUNCTION_COUNT: lambda _: Int64Type(),
    afpb.AGGREGATION_FUNCTION_LAST: lambda feature_type: feature_type,
    afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N: lambda feature_type: ArrayType(feature_type),
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: lambda feature_type: ArrayType(feature_type),
    afpb.AGGREGATION_FUNCTION_MEAN: lambda _: Float64Type(),
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: lambda _: Float64Type(),
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: lambda _: Float64Type(),
    afpb.AGGREGATION_FUNCTION_VAR_POP: lambda _: Float64Type(),
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: lambda _: Float64Type(),
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: lambda feature_type: ArrayType(feature_type),
    afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N: lambda feature_type: ArrayType(feature_type),
    afpb.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT: lambda _: Int64Type(),
    afpb.AGGREGATION_FUNCTION_APPROX_PERCENTILE: lambda _: Float64Type(),
}


@typechecked
def get_aggregation_function_result_type(
    aggregation_function_enum: afpb.AggregationFunction, feature_type: DataType
) -> DataType:
    return _AGGREGATION_RESULT_TYPE[aggregation_function_enum](feature_type)


@dataclass
class QueryWindowSpec:
    partition_cols: List[str]
    order_by_col: str
    range_start: Union[str, WindowFrameAnalyticFunction.Edge]
    range_end: Union[str, WindowFrameAnalyticFunction.Edge]


def get_simple_window_query(
    col: str,
    query_window_spec: QueryWindowSpec,
    analytic_function: Type[WindowFrameAnalyticFunction],
    ignore_nulls: bool = False,
) -> Term:
    expr = analytic_function(Field(col))
    if ignore_nulls:
        expr = expr.ignore_nulls()
    return (
        expr.over(*[Field(x) for x in query_window_spec.partition_cols])
        .orderby(*[Field(query_window_spec.order_by_col)])
        .range(query_window_spec.range_start, query_window_spec.range_end)
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/tecton_api_key__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auth import principal__client_pb2 as tecton__proto_dot_auth_dot_principal____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.tecton_proto/data/tecton_api_key__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a)tecton_proto/auth/principal__client.proto\x1a$tecton_proto/common/id__client.proto\"\xa5\x04\n\x0cTectonApiKey\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x1d\n\nhashed_key\x18\x02 \x01(\tR\thashedKey\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12\x1a\n\x08\x61rchived\x18\x04 \x01(\x08R\x08\x61rchived\x12\x1d\n\ncreated_by\x18\x05 \x01(\tR\tcreatedBy\x12!\n\x0cobscured_key\x18\x06 \x01(\tR\x0bobscuredKey\x12\x19\n\x08is_admin\x18\x07 \x01(\x08R\x07isAdmin\x12\x12\n\x04name\x18\x08 \x01(\tR\x04name\x12!\n\tis_active\x18\t \x01(\x08:\x04trueR\x08isActive\x12\x36\n\x07\x63reator\x18\n \x01(\x0b\x32\x1c.tecton_proto.auth.PrincipalR\x07\x63reator\x12\x39\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12*\n\x11is_system_managed\x18\x0c \x01(\x08R\x0fisSystemManaged\x12!\n\x0cis_temporary\x18\r \x01(\x08R\x0bisTemporary\x12\x39\n\nexpires_at\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.TimestampR\texpiresAtB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.tecton_api_key__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _TECTONAPIKEY._serialized_start=184
  _TECTONAPIKEY._serialized_end=733
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from datetime import timedelta
from typing import Tuple

import pendulum

from tecton_core import query_consts
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.builder import build_aggregated_time_range_validity_query
from tecton_core.query.builder import build_temporal_time_range_validity_query
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query_consts import valid_from
from tecton_core.query_consts import valid_to
from tecton_core.schema_derivation_utils import compute_features_schema_from_feature_definition
from tecton_core.time_utils import convert_to_effective_timestamp
from tecton_materialization.common.task_params import TimeInterval
from tecton_materialization.ray.delta import OfflineStoreParams
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_materialization.ray.nodes import TimeSpec
from tecton_proto.common import data_type__client_pb2 as data_type_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2


def get_feature_export_qt(
    fd: FeatureDefinitionWrapper, materialization_start_time: datetime, materialization_end_time: datetime
) -> Tuple[NodeRef, TimeInterval]:
    # read api uses effective timestamps for querying time ranges
    effective_start_time = datetime.fromtimestamp(
        convert_to_effective_timestamp(
            int(materialization_start_time.timestamp()),
            fd.batch_materialization_schedule.in_seconds(),
            fd.max_source_data_delay.in_seconds(),
        ),
        pendulum.tz.UTC,
    )
    effective_end_time = datetime.fromtimestamp(
        convert_to_effective_timestamp(
            int(materialization_end_time.timestamp()),
            fd.batch_materialization_schedule.in_seconds(),
            fd.max_source_data_delay.in_seconds(),
        ),
        pendulum.tz.UTC,
    )

    # TODO (TEC-18861): this is temporary pending getting tecton sdk into anyscale/ray cluster environment
    # will switch to `fv.get_features_in_range()` instead
    query_time_range = pendulum.Period(effective_start_time, effective_end_time)
    if fd.is_temporal:
        start_lookback = pendulum.from_timestamp(0)
        end_lookback = query_time_range.end + fd.batch_materialization_schedule
        if fd.serving_ttl:
            start_lookback = query_time_range.start - fd.serving_ttl - fd.max_source_data_delay
        if fd.feature_start_timestamp:
            start_lookback = max(start_lookback, fd.feature_start_timestamp)
        lookback_time_range = pendulum.Period(start_lookback, end_lookback)
        qt = build_temporal_time_range_validity_query(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            fd=fd,
            from_source=False,
            query_time_range=query_time_range,
            lookback_time_range=lookback_time_range,
            entities=None,
        )
    elif fd.is_temporal_aggregate:
        feature_data_start_limit = query_time_range.start - fd.max_source_data_delay
        feature_data_end_limit = query_time_range.end + fd.batch_materialization_schedule
        if fd.feature_start_timestamp:
            feature_data_start_limit = max(feature_data_start_limit, fd.feature_start_timestamp)
        if fd.has_lifetime_aggregate:
            feature_data_start_limit = pendulum.from_timestamp(0)
        else:
            feature_data_start_limit = feature_data_start_limit + fd.earliest_window_start
        if fd.materialization_start_timestamp:
            feature_data_start_limit = max(feature_data_start_limit, fd.materialization_start_timestamp)

        feature_data_time_limits = pendulum.Period(feature_data_start_limit, feature_data_end_limit)
        qt = build_aggregated_time_range_validity_query(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            fdw=fd,
            from_source=False,
            query_time_range=query_time_range,
            feature_data_time_limits=feature_data_time_limits,
            entities=None,
        )
    else:
        msg = "FV type not supported for feature export."
        raise ValueError(msg)

    # add partition column, partitioned on _valid_from timestamp
    qt = AddTimePartitionNode(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        input_node=qt,
        time_spec=feature_export_time_spec(),
    ).as_ref()

    # return the interval we queried
    interval = TimeInterval(effective_start_time, effective_end_time)

    return qt, interval


def get_feature_export_store_params(fd: FeatureDefinitionWrapper) -> OfflineStoreParams:
    schema = schema_pb2.Schema(columns=compute_features_schema_from_feature_definition(fd).columns)
    to_remove = [col for col in schema.columns if col.name in (fd.timestamp_key, query_consts.anchor_time())]
    for col in to_remove:
        schema.columns.remove(col)

    valid_from_column = schema_pb2.Column(
        name=valid_from(),
        offline_data_type=data_type_pb2.DataType(type=data_type_pb2.DataTypeEnum.DATA_TYPE_TIMESTAMP),
    )
    valid_to_column = schema_pb2.Column(
        name=valid_to(),
        offline_data_type=data_type_pb2.DataType(type=data_type_pb2.DataTypeEnum.DATA_TYPE_TIMESTAMP),
    )
    schema.columns.append(valid_from_column)
    schema.columns.append(valid_to_column)
    feature_params = OfflineStoreParams(
        feature_view_id=fd.id,
        feature_view_name=fd.name,
        schema=schema,
        time_spec=feature_export_time_spec(),
        feature_store_format_version=fd.get_feature_store_format_version,
        batch_schedule=fd.get_batch_schedule_for_version,
    )
    return feature_params


def feature_export_time_spec():
    return TimeSpec(
        timestamp_key=valid_from(),
        partition_size=timedelta(days=1),
        time_column=valid_from(),
        partition_is_anchor=False,
    )



import time
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional

from tecton._internals import metadata_service
from tecton._internals import utils
from tecton._internals.display import Displayable
from tecton_proto.common import id__client_pb2 as id_pb2
from tecton_proto.data.materialization_status__client_pb2 import MaterializationStatus
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import CancelJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetJobRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetLatestReadyTimeRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import GetLatestReadyTimeResponse
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import JobAttempt
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import ListJobsRequest
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import MaterializationJob
from tecton_proto.materializationjobservice.materialization_job_service__client_pb2 import MaterializationJobRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetMaterializationStatusRequest


class MaterializationTimeoutException(Exception):
    pass


class MaterializationJobFailedException(Exception):
    pass


@dataclass
class MaterializationAttemptData:
    """
    Data representation of the materialization job attempt.

    Materialization job may have multiple attempts to materialize features.

    :param id: ID string of the materialization attempt.
    :param run_url: URL to track materialization attempt.
    :param state: State of the materialization attempt.
    :param created_at: Materialization attempt creation timestamp.
    :param updated_at: Materialization attempt update timestamp.
    """

    id: str
    run_url: str
    state: str
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_proto(cls, proto: JobAttempt):
        created_at = datetime.utcfromtimestamp(proto.created_at.seconds)
        updated_at = datetime.utcfromtimestamp(proto.updated_at.seconds)
        return cls(id=proto.id, run_url=proto.run_url, state=proto.state, created_at=created_at, updated_at=updated_at)


@dataclass
class MaterializationJobData:
    """
    Data representation of the materialization job

    :param id: ID string of the materialization job.
    :param workspace: Name of the project workspace.
    :param feature_view: Name of the Feature View.
    :param state: State of the materialization job.
    :param online: Whether the job materializes features to the online store.
    :param offline: Whether the job materializes features to the offline store.
    :param start_time: Start timestamp of the batch materialization window.
    :param end_time: End timestamp of the batch materialization window.
    :param created_at: Job creation timestamp.
    :param updated_at: Job update timestamp.
    :param attempts: Materialization attempts. List of :class:`MaterializationAttemptData`
    :param next_attempt_at: If job needs another attempt, Start timestamp the next materialization attempt.
    :param job_type: Type of materialization. One of 'BATCH' or 'STREAM'.
    """

    id: str
    workspace: str
    feature_view: str
    state: str
    online: bool
    offline: bool
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    attempts: List[MaterializationAttemptData]
    next_attempt_at: Optional[datetime]
    job_type: str

    @classmethod
    def from_proto(cls, proto: MaterializationJob):
        start_time = datetime.utcfromtimestamp(proto.start_time.seconds) if proto.HasField("start_time") else None
        end_time = datetime.utcfromtimestamp(proto.end_time.seconds) if proto.HasField("end_time") else None
        created_at = datetime.utcfromtimestamp(proto.created_at.seconds)
        updated_at = datetime.utcfromtimestamp(proto.updated_at.seconds)
        attempts = [MaterializationAttemptData.from_proto(attempt_proto) for attempt_proto in proto.attempts]
        next_attempt_at = (
            datetime.utcfromtimestamp(proto.next_attempt_at.seconds) if proto.HasField("next_attempt_at") else None
        )
        return cls(
            id=proto.id,
            workspace=proto.workspace,
            feature_view=proto.feature_view,
            state=proto.state,
            online=proto.online,
            offline=proto.offline,
            start_time=start_time,
            end_time=end_time,
            created_at=created_at,
            updated_at=updated_at,
            attempts=attempts,
            next_attempt_at=next_attempt_at,
            job_type=proto.job_type,
        )


def trigger_materialization_job(
    feature_view: str,
    workspace: str,
    start_time: datetime,
    end_time: datetime,
    online: bool,
    offline: bool,
    use_tecton_managed_retries: bool = True,
    overwrite: bool = False,
) -> str:
    """
    Starts a batch materialization job for this Feature View.

    :param start_time: The job will materialize feature values between the start_time and end_time.
    :param end_time: The job will materialize feature values between the start_time and end_time.
    :param online: Whether the job will materialize features to the online store.
        The Feature View must be configured with online=True in order to materialize features online.
    :param offline: Whether the job will materialize features to the offline store.
        The Feature View must be configured with offline=True in order to materialize features offline.
    :param use_tecton_managed_retries: If enabled, Tecton will automatically retry failed attempts.
        Disable to manage your own retry behavior.
    :param overwrite: If enabled, you will be able to run materialization jobs for periods that previously have materialized data.
        This operation can be sensitive for feature views with existing materialized online data.
        For the offline store, all previously materialized data between the start time and end time will be dropped.
        For the online store, all previous data will remain, but may be overwritten by this job.
    :return: ID string of the created materialization job.
    :raises TectonValidationError: If job params are not valid.
    """
    request = MaterializationJobRequest()
    request.feature_view = feature_view
    request.workspace = workspace
    request.start_time.FromDatetime(start_time)
    request.end_time.FromDatetime(end_time)
    request.online = online
    request.offline = offline
    request.use_tecton_managed_retries = use_tecton_managed_retries
    request.overwrite = overwrite

    mds_instance = metadata_service.instance()
    response = mds_instance.SubmitMaterializationJob(request)
    return response.job.id


def list_materialization_jobs(feature_view: str, workspace: str) -> List[MaterializationJobData]:
    """
    Retrieves the list of all materialization jobs for this Feature View.

    :return: List of :class:`MaterializationJobData` objects.
    """
    request = ListJobsRequest()
    request.feature_view = feature_view
    request.workspace = workspace

    mds_instance = metadata_service.instance()
    response = mds_instance.ListMaterializationJobs(request)
    return [MaterializationJobData.from_proto(job) for job in response.jobs]


def get_materialization_job(feature_view: str, workspace: str, job_id: str) -> MaterializationJobData:
    """
    Retrieves data about the specified materialization job for this Feature View.

    This data includes information about job attempts.

    :param job_id: ID string of the materialization job.
    :return: :class:`MaterializationJobData` object for the job.
    """
    request = GetJobRequest()
    request.feature_view = feature_view
    request.workspace = workspace
    request.job_id = job_id

    mds_instance = metadata_service.instance()
    response = mds_instance.GetMaterializationJob(request)
    return MaterializationJobData.from_proto(response.job)


def get_latest_ready_time(feature_view: str, workspace: str) -> GetLatestReadyTimeResponse:
    request = GetLatestReadyTimeRequest()
    request.feature_view = feature_view
    request.workspace = workspace

    mds_instance = metadata_service.instance()
    return mds_instance.GetLatestReadyTime(request)


def cancel_materialization_job(feature_view: str, workspace: str, job_id: str) -> MaterializationJobData:
    """
    Cancels the scheduled or running batch materialization job for this Feature View specified by the job identifier.
    Once cancelled, a job will not be retried further.

    Job run state will be set to MANUAL_CANCELLATION_REQUESTED.
    Note that cancellation is asynchronous, so it may take some time for the cancellation to complete.
    If job run is already in MANUAL_CANCELLATION_REQUESTED or in a terminal state then it'll return the job.

    :param job_id: ID string of the materialization job.
    :return: :class:`MaterializationJobData` object for the cancelled job.
    """
    request = CancelJobRequest()
    request.feature_view = feature_view
    request.workspace = workspace
    request.job_id = job_id

    mds_instance = metadata_service.instance()
    response = mds_instance.CancelMaterializationJob(request)
    return MaterializationJobData.from_proto(response.job)


def wait_for_materialization_job(
    feature_view: str,
    workspace: str,
    job_id: str,
    timeout: Optional[timedelta] = None,
) -> MaterializationJobData:
    """
    Blocks until the specified job has been completed.

    :param job_id: ID string of the materialization job.
    :param timeout: (Optional) timeout for this function.
        An exception is raised if the job does not complete within the specified time.
    :return: :class:`MaterializationJobData` object for the successful job.
    :raises MaterializationTimeoutException:
        If timeout param is specified and job does not complete within the specified time.
    :raises MaterializationJobFailedException: If materialization job did not reach a successful state.
    """
    wait_start_time = datetime.now()
    while True:
        job_data = get_materialization_job(feature_view, workspace, job_id)
        run_state = job_data.state

        if run_state == "SUCCESS":
            return job_data
        elif timeout and ((datetime.now() - wait_start_time) > timeout):
            msg = f"job {job_id} timed out, last job state {run_state}"
            raise MaterializationTimeoutException(msg)
        elif run_state == "RUNNING":
            time.sleep(60)
        else:
            msg = f"job {job_id} failed, last job state {run_state}"
            raise MaterializationJobFailedException(msg)


def get_materialization_status_response(id_proto: id_pb2.Id, workspace: str) -> MaterializationStatus:
    """Returns MaterializationStatus proto for the FeatureView."""
    request = GetMaterializationStatusRequest()
    request.feature_package_id.CopyFrom(id_proto)
    request.workspace = workspace

    response = metadata_service.instance().GetMaterializationStatus(request)
    return response.materialization_status


def _create_materialization_table(column_names: List[str], materialization_status_rows: List[List]) -> Displayable:
    # Setting `max_width=0` creates a table with an unlimited width.
    table = Displayable.from_table(headings=column_names, rows=materialization_status_rows, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(column_names))])

    return table


def get_materialization_status_for_display(
    id_proto: id_pb2.Id, workspace: str, verbose: bool, limit: int, sort_columns: Optional[str], errors_only: bool
) -> Displayable:
    materialization_attempts = get_materialization_status_response(id_proto, workspace).materialization_attempts
    column_names, materialization_status_rows = utils.format_materialization_attempts(
        materialization_attempts, verbose, limit, sort_columns, errors_only
    )

    return _create_materialization_table(column_names, materialization_status_rows)



import sys

import click

from tecton._internals import metadata_service
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


@click.group("user", cls=TectonGroup)
def user():
    """Manage users."""


@user.command("invite", help="Invite users to Tecton cluster")
@click.option("-u", "--user", default=None, help="User email")
@click.option("-f", "--file", default=None, help="File containing user emails, newline-separated", type=click.File("r"))
def invite(user, file):
    _process_user_action(user, file, _invite_user)


@user.command("delete", help="Delete users from Tecton cluster")
@click.option("-u", "--user", default=None, help="User email")
@click.option("-f", "--file", default=None, help="File containing user emails, newline-separated", type=click.File("r"))
def delete(user, file):
    _process_user_action(user, file, _delete_user)


def _process_user_action(user, file, handler):
    if file is not None:
        if user:
            msg = "Please use exactly one of --user or --file"
            raise click.BadArgumentUsage(msg)

        for u in [line.strip() for line in file.readlines() if len(line.strip()) > 0]:
            handler(u)
    elif user is not None:
        handler(user)
    else:
        msg = "Please submit one of --user or --file."
        raise click.BadArgumentUsage(msg)


def _invite_user(user_email: str) -> None:
    try:
        request = metadata_service_pb2.CreateClusterUserRequest(login_email=user_email)
        metadata_service.instance().CreateClusterUser(request)
    except Exception as e:
        printer.safe_print(f"Failed to invite [{user_email}]: {e}", file=sys.stderr)
        sys.exit(1)
    printer.safe_print(f"Successfully invited [{user_email}]")


def _delete_user(user_email: str) -> None:
    try:
        get_user_request = metadata_service_pb2.GetUserRequest(email=user_email)
        response = metadata_service.instance().GetUser(get_user_request)
        okta_id = response.user.okta_id

        delete_request = metadata_service_pb2.DeleteClusterUserRequest(okta_id=okta_id)
        metadata_service.instance().DeleteClusterUser(delete_request)
    except Exception as e:
        printer.safe_print(f"Failed to delete [{user_email}]: {e}", file=sys.stderr)
        sys.exit(1)
    printer.safe_print(f"Successfully deleted [{user_email}]")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/materialization/materialization_states__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nAtecton_proto/materialization/materialization_states__client.proto\x12\x1ctecton_proto.materialization*\xc0\x03\n\x18MaterializationTaskState\x12+\n\'MATERIALIZATION_TASK_STATUS_UNSPECIFIED\x10\x00\x12\'\n#MATERIALIZATION_TASK_STATUS_RUNNING\x10\x01\x12(\n$MATERIALIZATION_TASK_STATUS_DRAINING\x10\x02\x12\'\n#MATERIALIZATION_TASK_STATUS_DRAINED\x10\x05\x12=\n9MATERIALIZATION_TASK_STATUS_MANUAL_CANCELLATION_REQUESTED\x10\x07\x12\x32\n.MATERIALIZATION_TASK_STATUS_MANUALLY_CANCELLED\x10\x08\x12\'\n#MATERIALIZATION_TASK_STATUS_SUCCESS\x10\x03\x12\x31\n-MATERIALIZATION_TASK_STATUS_PERMANENT_FAILURE\x10\x04\x12,\n(MATERIALIZATION_TASK_STATUS_MANUAL_RETRY\x10\x06*\xad\x02\n\x1fMaterializationTaskAttemptState\x12.\n*MATERIALIZATION_ATTEMPT_STATUS_UNSPECIFIED\x10\x00\x12*\n&MATERIALIZATION_ATTEMPT_STATUS_PENDING\x10\x01\x12*\n&MATERIALIZATION_ATTEMPT_STATUS_RUNNING\x10\x02\x12*\n&MATERIALIZATION_ATTEMPT_STATUS_SUCCESS\x10\x03\x12(\n$MATERIALIZATION_ATTEMPT_STATUS_ERROR\x10\x04\x12,\n(MATERIALIZATION_ATTEMPT_STATUS_CANCELLED\x10\x05*\xa5\x01\n\x1aOnlineStoreCopierTaskState\x12\x1d\n\x19\x43OPIER_STATUS_UNSPECIFIED\x10\x00\x12\x19\n\x15\x43OPIER_STATUS_PENDING\x10\x01\x12\x19\n\x15\x43OPIER_STATUS_RUNNING\x10\x02\x12\x19\n\x15\x43OPIER_STATUS_SUCCESS\x10\x03\x12\x17\n\x13\x43OPIER_STATUS_ERROR\x10\x04\x42\x1e\n\x1a\x63om.tecton.materializationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.materialization.materialization_states__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.materializationP\001'
  _MATERIALIZATIONTASKSTATE._serialized_start=100
  _MATERIALIZATIONTASKSTATE._serialized_end=548
  _MATERIALIZATIONTASKATTEMPTSTATE._serialized_start=551
  _MATERIALIZATIONTASKATTEMPTSTATE._serialized_end=852
  _ONLINESTORECOPIERTASKSTATE._serialized_start=855
  _ONLINESTORECOPIERTASKSTATE._serialized_end=1020
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from py4j.java_gateway import JavaClass

from pyspark import RDD, since
from pyspark.sql.column import _to_seq, _to_java_column
from pyspark.sql.types import StructType
from pyspark.sql import utils
from pyspark.sql.utils import to_str

__all__ = ["DataFrameReader", "DataFrameWriter"]


class OptionUtils(object):

    def _set_opts(self, schema=None, **options):
        """
        Set named options (filter out those the value is None)
        """
        if schema is not None:
            self.schema(schema)
        for k, v in options.items():
            if v is not None:
                self.option(k, v)


class DataFrameReader(OptionUtils):
    """
    Interface used to load a :class:`DataFrame` from external storage systems
    (e.g. file systems, key-value stores, etc). Use :attr:`SparkSession.read`
    to access this.

    .. versionadded:: 1.4
    """

    def __init__(self, spark):
        self._jreader = spark._ssql_ctx.read()
        self._spark = spark

    def _df(self, jdf):
        from pyspark.sql.dataframe import DataFrame
        return DataFrame(jdf, self._spark)

    def format(self, source):
        """Specifies the input data source format.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        source : str
            string, name of the data source, e.g. 'json', 'parquet'.

        Examples
        --------
        >>> df = spark.read.format('json').load('python/test_support/sql/people.json')
        >>> df.dtypes
        [('age', 'bigint'), ('name', 'string')]

        """
        self._jreader = self._jreader.format(source)
        return self

    def schema(self, schema):
        """Specifies the input schema.

        Some data sources (e.g. JSON) can infer the input schema automatically from data.
        By specifying the schema here, the underlying data source can skip the schema
        inference step, and thus speed up data loading.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        schema : :class:`pyspark.sql.types.StructType` or str
            a :class:`pyspark.sql.types.StructType` object or a DDL-formatted string
            (For example ``col0 INT, col1 DOUBLE``).

        >>> s = spark.read.schema("col0 INT, col1 DOUBLE")
        """
        from pyspark.sql import SparkSession
        spark = SparkSession.builder.getOrCreate()
        if isinstance(schema, StructType):
            jschema = spark._jsparkSession.parseDataType(schema.json())
            self._jreader = self._jreader.schema(jschema)
        elif isinstance(schema, str):
            self._jreader = self._jreader.schema(schema)
        else:
            raise TypeError("schema should be StructType or string")
        return self

    @since(1.5)
    def option(self, key, value):
        """Adds an input option for the underlying data source.

        You can set the following option(s) for reading files:
            * ``timeZone``: sets the string that indicates a time zone ID to be used to parse
                timestamps in the JSON/CSV datasources or partition values. The following
                formats of `timeZone` are supported:

                * Region-based zone ID: It should have the form 'area/city', such as \
                  'America/Los_Angeles'.
                * Zone offset: It should be in the format '(+|-)HH:mm', for example '-08:00' or \
                 '+01:00'. Also 'UTC' and 'Z' are supported as aliases of '+00:00'.

                Other short names like 'CST' are not recommended to use because they can be
                ambiguous. If it isn't set, the current value of the SQL config
                ``spark.sql.session.timeZone`` is used by default.
            * ``pathGlobFilter``: an optional glob pattern to only include files with paths matching
                the pattern. The syntax follows org.apache.hadoop.fs.GlobFilter.
                It does not change the behavior of partition discovery.
            * ``modifiedBefore``: an optional timestamp to only include files with
                modification times occurring before the specified time. The provided timestamp
                must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
            * ``modifiedAfter``: an optional timestamp to only include files with
                modification times occurring after the specified time. The provided timestamp
                must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        """
        self._jreader = self._jreader.option(key, to_str(value))
        return self

    @since(1.4)
    def options(self, **options):
        """Adds input options for the underlying data source.

        You can set the following option(s) for reading files:
            * ``timeZone``: sets the string that indicates a time zone ID to be used to parse
                timestamps in the JSON/CSV datasources or partition values. The following
                formats of `timeZone` are supported:

                * Region-based zone ID: It should have the form 'area/city', such as \
                  'America/Los_Angeles'.
                * Zone offset: It should be in the format '(+|-)HH:mm', for example '-08:00' or \
                 '+01:00'. Also 'UTC' and 'Z' are supported as aliases of '+00:00'.

                Other short names like 'CST' are not recommended to use because they can be
                ambiguous. If it isn't set, the current value of the SQL config
                ``spark.sql.session.timeZone`` is used by default.
            * ``pathGlobFilter``: an optional glob pattern to only include files with paths matching
                the pattern. The syntax follows org.apache.hadoop.fs.GlobFilter.
                It does not change the behavior of partition discovery.
            * ``modifiedBefore``: an optional timestamp to only include files with
                modification times occurring before the specified time. The provided timestamp
                must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
            * ``modifiedAfter``: an optional timestamp to only include files with
                modification times occurring after the specified time. The provided timestamp
                must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        """
        for k in options:
            self._jreader = self._jreader.option(k, to_str(options[k]))
        return self

    def load(self, path=None, format=None, schema=None, **options):
        """Loads data from a data source and returns it as a :class:`DataFrame`.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : str or list, optional
            optional string or a list of string for file-system backed data sources.
        format : str, optional
            optional string for format of the data source. Default to 'parquet'.
        schema : :class:`pyspark.sql.types.StructType` or str, optional
            optional :class:`pyspark.sql.types.StructType` for the input schema
            or a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).
        **options : dict
            all other string options

        Examples
        --------
        >>> df = spark.read.format("parquet").load('python/test_support/sql/parquet_partitioned',
        ...     opt1=True, opt2=1, opt3='str')
        >>> df.dtypes
        [('name', 'string'), ('year', 'int'), ('month', 'int'), ('day', 'int')]

        >>> df = spark.read.format('json').load(['python/test_support/sql/people.json',
        ...     'python/test_support/sql/people1.json'])
        >>> df.dtypes
        [('age', 'bigint'), ('aka', 'string'), ('name', 'string')]
        """
        if format is not None:
            self.format(format)
        if schema is not None:
            self.schema(schema)
        self.options(**options)
        if isinstance(path, str):
            return self._df(self._jreader.load(path))
        elif path is not None:
            if type(path) != list:
                path = [path]
            return self._df(self._jreader.load(self._spark._sc._jvm.PythonUtils.toSeq(path)))
        else:
            return self._df(self._jreader.load())

    def json(self, path, schema=None, primitivesAsString=None, prefersDecimal=None,
             allowComments=None, allowUnquotedFieldNames=None, allowSingleQuotes=None,
             allowNumericLeadingZero=None, allowBackslashEscapingAnyCharacter=None,
             mode=None, columnNameOfCorruptRecord=None, dateFormat=None, timestampFormat=None,
             multiLine=None, allowUnquotedControlChars=None, lineSep=None, samplingRatio=None,
             dropFieldIfAllNull=None, encoding=None, locale=None, pathGlobFilter=None,
             recursiveFileLookup=None, allowNonNumericNumbers=None,
             modifiedBefore=None, modifiedAfter=None):
        """
        Loads JSON files and returns the results as a :class:`DataFrame`.

        `JSON Lines <http://jsonlines.org/>`_ (newline-delimited JSON) is supported by default.
        For JSON (one record per file), set the ``multiLine`` parameter to ``true``.

        If the ``schema`` parameter is not specified, this function goes
        through the input once to determine the input schema.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : str, list or :class:`RDD`
            string represents path to the JSON dataset, or a list of paths,
            or RDD of Strings storing JSON objects.
        schema : :class:`pyspark.sql.types.StructType` or str, optional
            an optional :class:`pyspark.sql.types.StructType` for the input schema or
            a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).
        primitivesAsString : str or bool, optional
            infers all primitive values as a string type. If None is set,
            it uses the default value, ``false``.
        prefersDecimal : str or bool, optional
            infers all floating-point values as a decimal type. If the values
            do not fit in decimal, then it infers them as doubles. If None is
            set, it uses the default value, ``false``.
        allowComments : str or bool, optional
            ignores Java/C++ style comment in JSON records. If None is set,
            it uses the default value, ``false``.
        allowUnquotedFieldNames : str or bool, optional
            allows unquoted JSON field names. If None is set,
            it uses the default value, ``false``.
        allowSingleQuotes : str or bool, optional
            allows single quotes in addition to double quotes. If None is
            set, it uses the default value, ``true``.
        allowNumericLeadingZero : str or bool, optional
            allows leading zeros in numbers (e.g. 00012). If None is
            set, it uses the default value, ``false``.
        allowBackslashEscapingAnyCharacter : str or bool, optional
            allows accepting quoting of all character
            using backslash quoting mechanism. If None is
            set, it uses the default value, ``false``.
        mode : str, optional
            allows a mode for dealing with corrupt records during parsing. If None is
                     set, it uses the default value, ``PERMISSIVE``.

            * ``PERMISSIVE``: when it meets a corrupted record, puts the malformed string \
              into a field configured by ``columnNameOfCorruptRecord``, and sets malformed \
              fields to ``null``. To keep corrupt records, an user can set a string type \
              field named ``columnNameOfCorruptRecord`` in an user-defined schema. If a \
              schema does not have the field, it drops corrupt records during parsing. \
              When inferring a schema, it implicitly adds a ``columnNameOfCorruptRecord`` \
              field in an output schema.
            *  ``DROPMALFORMED``: ignores the whole corrupted records.
            *  ``FAILFAST``: throws an exception when it meets corrupted records.

        columnNameOfCorruptRecord: str, optional
            allows renaming the new field having malformed string
            created by ``PERMISSIVE`` mode. This overrides
            ``spark.sql.columnNameOfCorruptRecord``. If None is set,
            it uses the value specified in
            ``spark.sql.columnNameOfCorruptRecord``.
        dateFormat : str, optional
            sets the string that indicates a date format. Custom date formats
            follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to date type. If None is set, it uses the
            default value, ``yyyy-MM-dd``.
        timestampFormat : str, optional
            sets the string that indicates a timestamp format.
            Custom date formats follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to timestamp type. If None is set, it uses the
            default value, ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
        multiLine : str or bool, optional
            parse one record, which may span multiple lines, per file. If None is
            set, it uses the default value, ``false``.
        allowUnquotedControlChars : str or bool, optional
            allows JSON Strings to contain unquoted control
            characters (ASCII characters with value less than 32,
            including tab and line feed characters) or not.
        encoding : str or bool, optional
            allows to forcibly set one of standard basic or extended encoding for
            the JSON files. For example UTF-16BE, UTF-32LE. If None is set,
            the encoding of input JSON will be detected automatically
            when the multiLine option is set to ``true``.
        lineSep : str, optional
            defines the line separator that should be used for parsing. If None is
            set, it covers all ``\\r``, ``\\r\\n`` and ``\\n``.
        samplingRatio : str or float, optional
            defines fraction of input JSON objects used for schema inferring.
            If None is set, it uses the default value, ``1.0``.
        dropFieldIfAllNull : str or bool, optional
            whether to ignore column of all null values or empty
            array/struct during schema inference. If None is set, it
            uses the default value, ``false``.
        locale : str, optional
            sets a locale as language tag in IETF BCP 47 format. If None is set,
            it uses the default value, ``en-US``. For instance, ``locale`` is used while
            parsing dates and timestamps.
        pathGlobFilter : str or bool, optional
            an optional glob pattern to only include files with paths matching
            the pattern. The syntax follows `org.apache.hadoop.fs.GlobFilter`.
            It does not change the behavior of
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        recursiveFileLookup : str or bool, optional
            recursively scan a directory for files. Using this option
            disables
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        allowNonNumericNumbers : str or bool
            allows JSON parser to recognize set of "Not-a-Number" (NaN)
            tokens as legal floating number values. If None is set,
            it uses the default value, ``true``.

                * ``+INF``: for positive infinity, as well as alias of
                            ``+Infinity`` and ``Infinity``.
                *  ``-INF``: for negative infinity, alias ``-Infinity``.
                *  ``NaN``: for other not-a-numbers, like result of division by zero.
        modifiedBefore : an optional timestamp to only include files with
            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedAfter : an optional timestamp to only include files with
            modification times occurring after the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)


        Examples
        --------
        >>> df1 = spark.read.json('python/test_support/sql/people.json')
        >>> df1.dtypes
        [('age', 'bigint'), ('name', 'string')]
        >>> rdd = sc.textFile('python/test_support/sql/people.json')
        >>> df2 = spark.read.json(rdd)
        >>> df2.dtypes
        [('age', 'bigint'), ('name', 'string')]

        """
        self._set_opts(
            schema=schema, primitivesAsString=primitivesAsString, prefersDecimal=prefersDecimal,
            allowComments=allowComments, allowUnquotedFieldNames=allowUnquotedFieldNames,
            allowSingleQuotes=allowSingleQuotes, allowNumericLeadingZero=allowNumericLeadingZero,
            allowBackslashEscapingAnyCharacter=allowBackslashEscapingAnyCharacter,
            mode=mode, columnNameOfCorruptRecord=columnNameOfCorruptRecord, dateFormat=dateFormat,
            timestampFormat=timestampFormat, multiLine=multiLine,
            allowUnquotedControlChars=allowUnquotedControlChars, lineSep=lineSep,
            samplingRatio=samplingRatio, dropFieldIfAllNull=dropFieldIfAllNull, encoding=encoding,
            locale=locale, pathGlobFilter=pathGlobFilter, recursiveFileLookup=recursiveFileLookup,
            modifiedBefore=modifiedBefore, modifiedAfter=modifiedAfter,
            allowNonNumericNumbers=allowNonNumericNumbers)
        if isinstance(path, str):
            path = [path]
        if type(path) == list:
            return self._df(self._jreader.json(self._spark._sc._jvm.PythonUtils.toSeq(path)))
        elif isinstance(path, RDD):
            def func(iterator):
                for x in iterator:
                    if not isinstance(x, str):
                        x = str(x)
                    if isinstance(x, str):
                        x = x.encode("utf-8")
                    yield x
            keyed = path.mapPartitions(func)
            keyed._bypass_serializer = True
            jrdd = keyed._jrdd.map(self._spark._jvm.BytesToString())
            return self._df(self._jreader.json(jrdd))
        else:
            raise TypeError("path can be only string, list or RDD")

    def table(self, tableName):
        """Returns the specified table as a :class:`DataFrame`.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        tableName : str
            string, name of the table.

        Examples
        --------
        >>> df = spark.read.parquet('python/test_support/sql/parquet_partitioned')
        >>> df.createOrReplaceTempView('tmpTable')
        >>> spark.read.table('tmpTable').dtypes
        [('name', 'string'), ('year', 'int'), ('month', 'int'), ('day', 'int')]
        """
        return self._df(self._jreader.table(tableName))

    def parquet(self, *paths, **options):
        """
        Loads Parquet files, returning the result as a :class:`DataFrame`.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        paths : str

        Other Parameters
        ----------------
        mergeSchema : str or bool, optional
            sets whether we should merge schemas collected from all
            Parquet part-files. This will override
            ``spark.sql.parquet.mergeSchema``. The default value is specified in
            ``spark.sql.parquet.mergeSchema``.
        pathGlobFilter : str or bool, optional
            an optional glob pattern to only include files with paths matching
            the pattern. The syntax follows `org.apache.hadoop.fs.GlobFilter`.
            It does not change the behavior of
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        recursiveFileLookup : str or bool, optional
            recursively scan a directory for files. Using this option
            disables
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa

            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedBefore (batch only) : an optional timestamp to only include files with
            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedAfter (batch only) : an optional timestamp to only include files with
            modification times occurring after the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)

        Examples
        --------
        >>> df = spark.read.parquet('python/test_support/sql/parquet_partitioned')
        >>> df.dtypes
        [('name', 'string'), ('year', 'int'), ('month', 'int'), ('day', 'int')]
        """
        mergeSchema = options.get('mergeSchema', None)
        pathGlobFilter = options.get('pathGlobFilter', None)
        modifiedBefore = options.get('modifiedBefore', None)
        modifiedAfter = options.get('modifiedAfter', None)
        recursiveFileLookup = options.get('recursiveFileLookup', None)
        self._set_opts(mergeSchema=mergeSchema, pathGlobFilter=pathGlobFilter,
                       recursiveFileLookup=recursiveFileLookup, modifiedBefore=modifiedBefore,
                       modifiedAfter=modifiedAfter)

        return self._df(self._jreader.parquet(_to_seq(self._spark._sc, paths)))

    def text(self, paths, wholetext=False, lineSep=None, pathGlobFilter=None,
             recursiveFileLookup=None, modifiedBefore=None,
             modifiedAfter=None):
        """
        Loads text files and returns a :class:`DataFrame` whose schema starts with a
        string column named "value", and followed by partitioned columns if there
        are any.
        The text files must be encoded as UTF-8.

        By default, each line in the text file is a new row in the resulting DataFrame.

        .. versionadded:: 1.6.0

        Parameters
        ----------
        paths : str or list
            string, or list of strings, for input path(s).
        wholetext : str or bool, optional
            if true, read each file from input path(s) as a single row.
        lineSep : str, optional
            defines the line separator that should be used for parsing. If None is
            set, it covers all ``\\r``, ``\\r\\n`` and ``\\n``.
        pathGlobFilter : str or bool, optional
            an optional glob pattern to only include files with paths matching
            the pattern. The syntax follows `org.apache.hadoop.fs.GlobFilter`.
            It does not change the behavior of
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        recursiveFileLookup : str or bool, optional
            recursively scan a directory for files. Using this option disables
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa

            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedBefore (batch only) : an optional timestamp to only include files with
            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedAfter (batch only) : an optional timestamp to only include files with
            modification times occurring after the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)

        Examples
        --------
        >>> df = spark.read.text('python/test_support/sql/text-test.txt')
        >>> df.collect()
        [Row(value='hello'), Row(value='this')]
        >>> df = spark.read.text('python/test_support/sql/text-test.txt', wholetext=True)
        >>> df.collect()
        [Row(value='hello\\nthis')]
        """
        self._set_opts(
            wholetext=wholetext, lineSep=lineSep, pathGlobFilter=pathGlobFilter,
            recursiveFileLookup=recursiveFileLookup, modifiedBefore=modifiedBefore,
            modifiedAfter=modifiedAfter)

        if isinstance(paths, str):
            paths = [paths]
        return self._df(self._jreader.text(self._spark._sc._jvm.PythonUtils.toSeq(paths)))

    def csv(self, path, schema=None, sep=None, encoding=None, quote=None, escape=None,
            comment=None, header=None, inferSchema=None, ignoreLeadingWhiteSpace=None,
            ignoreTrailingWhiteSpace=None, nullValue=None, nanValue=None, positiveInf=None,
            negativeInf=None, dateFormat=None, timestampFormat=None, maxColumns=None,
            maxCharsPerColumn=None, maxMalformedLogPerPartition=None, mode=None,
            columnNameOfCorruptRecord=None, multiLine=None, charToEscapeQuoteEscaping=None,
            samplingRatio=None, enforceSchema=None, emptyValue=None, locale=None, lineSep=None,
            pathGlobFilter=None, recursiveFileLookup=None, modifiedBefore=None, modifiedAfter=None,
            unescapedQuoteHandling=None):
        r"""Loads a CSV file and returns the result as a  :class:`DataFrame`.

        This function will go through the input once to determine the input schema if
        ``inferSchema`` is enabled. To avoid going through the entire data once, disable
        ``inferSchema`` option or specify the schema explicitly using ``schema``.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        path : str or list
            string, or list of strings, for input path(s),
            or RDD of Strings storing CSV rows.
        schema : :class:`pyspark.sql.types.StructType` or str, optional
            an optional :class:`pyspark.sql.types.StructType` for the input schema
            or a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).
        sep : str, optional
            sets a separator (one or more characters) for each field and value. If None is
            set, it uses the default value, ``,``.
        encoding : str, optional
            decodes the CSV files by the given encoding type. If None is set,
            it uses the default value, ``UTF-8``.
        quote : str, optional
            sets a single character used for escaping quoted values where the
            separator can be part of the value. If None is set, it uses the default
            value, ``"``. If you would like to turn off quotations, you need to set an
            empty string.
        escape : str, optional
            sets a single character used for escaping quotes inside an already
            quoted value. If None is set, it uses the default value, ``\``.
        comment : str, optional
            sets a single character used for skipping lines beginning with this
            character. By default (None), it is disabled.
        header : str or bool, optional
            uses the first line as names of columns. If None is set, it uses the
            default value, ``false``.

            .. note:: if the given path is a RDD of Strings, this header
                option will remove all lines same with the header if exists.

        inferSchema : str or bool, optional
            infers the input schema automatically from data. It requires one extra
            pass over the data. If None is set, it uses the default value, ``false``.
        enforceSchema : str or bool, optional
            If it is set to ``true``, the specified or inferred schema will be
            forcibly applied to datasource files, and headers in CSV files will be
            ignored. If the option is set to ``false``, the schema will be
            validated against all headers in CSV files or the first header in RDD
            if the ``header`` option is set to ``true``. Field names in the schema
            and column names in CSV headers are checked by their positions
            taking into account ``spark.sql.caseSensitive``. If None is set,
            ``true`` is used by default. Though the default value is ``true``,
            it is recommended to disable the ``enforceSchema`` option
            to avoid incorrect results.
        ignoreLeadingWhiteSpace : str or bool, optional
            A flag indicating whether or not leading whitespaces from
            values being read should be skipped. If None is set, it
            uses the default value, ``false``.
        ignoreTrailingWhiteSpace : str or bool, optional
            A flag indicating whether or not trailing whitespaces from
            values being read should be skipped. If None is set, it
            uses the default value, ``false``.
        nullValue : str, optional
            sets the string representation of a null value. If None is set, it uses
            the default value, empty string. Since 2.0.1, this ``nullValue`` param
            applies to all supported types including the string type.
        nanValue : str, optional
            sets the string representation of a non-number value. If None is set, it
            uses the default value, ``NaN``.
        positiveInf : str, optional
            sets the string representation of a positive infinity value. If None
            is set, it uses the default value, ``Inf``.
        negativeInf : str, optional
            sets the string representation of a negative infinity value. If None
            is set, it uses the default value, ``Inf``.
        dateFormat : str, optional
            sets the string that indicates a date format. Custom date formats
            follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to date type. If None is set, it uses the
            default value, ``yyyy-MM-dd``.
        timestampFormat : str, optional
            sets the string that indicates a timestamp format.
            Custom date formats follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to timestamp type. If None is set, it uses the
            default value, ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
        maxColumns : str or int, optional
            defines a hard limit of how many columns a record can have. If None is
            set, it uses the default value, ``20480``.
        maxCharsPerColumn : str or int, optional
            defines the maximum number of characters allowed for any given
            value being read. If None is set, it uses the default value,
            ``-1`` meaning unlimited length.
        maxMalformedLogPerPartition : str or int, optional
            this parameter is no longer used since Spark 2.2.0.
            If specified, it is ignored.
        mode : str, optional
            allows a mode for dealing with corrupt records during parsing. If None is
            set, it uses the default value, ``PERMISSIVE``. Note that Spark tries to
            parse only required columns in CSV under column pruning. Therefore, corrupt
            records can be different based on required set of fields. This behavior can
            be controlled by ``spark.sql.csv.parser.columnPruning.enabled``
            (enabled by default).

            * ``PERMISSIVE``: when it meets a corrupted record, puts the malformed string \
              into a field configured by ``columnNameOfCorruptRecord``, and sets malformed \
              fields to ``null``. To keep corrupt records, an user can set a string type \
              field named ``columnNameOfCorruptRecord`` in an user-defined schema. If a \
              schema does not have the field, it drops corrupt records during parsing. \
              A record with less/more tokens than schema is not a corrupted record to CSV. \
              When it meets a record having fewer tokens than the length of the schema, \
              sets ``null`` to extra fields. When the record has more tokens than the \
              length of the schema, it drops extra tokens.
            * ``DROPMALFORMED``: ignores the whole corrupted records.
            * ``FAILFAST``: throws an exception when it meets corrupted records.

        columnNameOfCorruptRecord : str, optional
            allows renaming the new field having malformed string
            created by ``PERMISSIVE`` mode. This overrides
            ``spark.sql.columnNameOfCorruptRecord``. If None is set,
            it uses the value specified in
            ``spark.sql.columnNameOfCorruptRecord``.
        multiLine : str or bool, optional
            parse records, which may span multiple lines. If None is
            set, it uses the default value, ``false``.
        charToEscapeQuoteEscaping : str, optional
            sets a single character used for escaping the escape for
            the quote character. If None is set, the default value is
            escape character when escape and quote characters are
            different, ``\0`` otherwise.
        samplingRatio : str or float, optional
            defines fraction of rows used for schema inferring.
            If None is set, it uses the default value, ``1.0``.
        emptyValue : str, optional
            sets the string representation of an empty value. If None is set, it uses
            the default value, empty string.
        locale : str, optional
            sets a locale as language tag in IETF BCP 47 format. If None is set,
            it uses the default value, ``en-US``. For instance, ``locale`` is used while
            parsing dates and timestamps.
        lineSep : str, optional
            defines the line separator that should be used for parsing. If None is
            set, it covers all ``\\r``, ``\\r\\n`` and ``\\n``.
            Maximum length is 1 character.
        pathGlobFilter : str or bool, optional
            an optional glob pattern to only include files with paths matching
            the pattern. The syntax follows `org.apache.hadoop.fs.GlobFilter`.
            It does not change the behavior of
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        recursiveFileLookup : str or bool, optional
            recursively scan a directory for files. Using this option disables
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa

            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedBefore (batch only) : an optional timestamp to only include files with
            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedAfter (batch only) : an optional timestamp to only include files with
            modification times occurring after the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        unescapedQuoteHandling : str, optional
            defines how the CsvParser will handle values with unescaped quotes. If None is
            set, it uses the default value, ``STOP_AT_DELIMITER``.

            * ``STOP_AT_CLOSING_QUOTE``: If unescaped quotes are found in the input, accumulate
              the quote character and proceed parsing the value as a quoted value, until a closing
              quote is found.
            * ``BACK_TO_DELIMITER``: If unescaped quotes are found in the input, consider the value
              as an unquoted value. This will make the parser accumulate all characters of the current
              parsed value until the delimiter is found. If no delimiter is found in the value, the
              parser will continue accumulating characters from the input until a delimiter or line
              ending is found.
            * ``STOP_AT_DELIMITER``: If unescaped quotes are found in the input, consider the value
              as an unquoted value. This will make the parser accumulate all characters until the
              delimiter or a line ending is found in the input.
            * ``SKIP_VALUE``: If unescaped quotes are found in the input, the content parsed
              for the given value will be skipped and the value set in nullValue will be produced
              instead.
            * ``RAISE_ERROR``: If unescaped quotes are found in the input, a TextParsingException
              will be thrown.

        Examples
        --------
        >>> df = spark.read.csv('python/test_support/sql/ages.csv')
        >>> df.dtypes
        [('_c0', 'string'), ('_c1', 'string')]
        >>> rdd = sc.textFile('python/test_support/sql/ages.csv')
        >>> df2 = spark.read.csv(rdd)
        >>> df2.dtypes
        [('_c0', 'string'), ('_c1', 'string')]
        """
        self._set_opts(
            schema=schema, sep=sep, encoding=encoding, quote=quote, escape=escape, comment=comment,
            header=header, inferSchema=inferSchema, ignoreLeadingWhiteSpace=ignoreLeadingWhiteSpace,
            ignoreTrailingWhiteSpace=ignoreTrailingWhiteSpace, nullValue=nullValue,
            nanValue=nanValue, positiveInf=positiveInf, negativeInf=negativeInf,
            dateFormat=dateFormat, timestampFormat=timestampFormat, maxColumns=maxColumns,
            maxCharsPerColumn=maxCharsPerColumn,
            maxMalformedLogPerPartition=maxMalformedLogPerPartition, mode=mode,
            columnNameOfCorruptRecord=columnNameOfCorruptRecord, multiLine=multiLine,
            charToEscapeQuoteEscaping=charToEscapeQuoteEscaping, samplingRatio=samplingRatio,
            enforceSchema=enforceSchema, emptyValue=emptyValue, locale=locale, lineSep=lineSep,
            pathGlobFilter=pathGlobFilter, recursiveFileLookup=recursiveFileLookup,
            modifiedBefore=modifiedBefore, modifiedAfter=modifiedAfter,
            unescapedQuoteHandling=unescapedQuoteHandling)
        if isinstance(path, str):
            path = [path]
        if type(path) == list:
            return self._df(self._jreader.csv(self._spark._sc._jvm.PythonUtils.toSeq(path)))
        elif isinstance(path, RDD):
            def func(iterator):
                for x in iterator:
                    if not isinstance(x, str):
                        x = str(x)
                    if isinstance(x, str):
                        x = x.encode("utf-8")
                    yield x
            keyed = path.mapPartitions(func)
            keyed._bypass_serializer = True
            jrdd = keyed._jrdd.map(self._spark._jvm.BytesToString())
            # see SPARK-22112
            # There aren't any jvm api for creating a dataframe from rdd storing csv.
            # We can do it through creating a jvm dataset firstly and using the jvm api
            # for creating a dataframe from dataset storing csv.
            jdataset = self._spark._ssql_ctx.createDataset(
                jrdd.rdd(),
                self._spark._jvm.Encoders.STRING())
            return self._df(self._jreader.csv(jdataset))
        else:
            raise TypeError("path can be only string, list or RDD")

    def orc(self, path, mergeSchema=None, pathGlobFilter=None, recursiveFileLookup=None,
            modifiedBefore=None, modifiedAfter=None):
        """Loads ORC files, returning the result as a :class:`DataFrame`.

        .. versionadded:: 1.5.0

        Parameters
        ----------
        path : str or list
        mergeSchema : str or bool, optional
            sets whether we should merge schemas collected from all
            ORC part-files. This will override ``spark.sql.orc.mergeSchema``.
            The default value is specified in ``spark.sql.orc.mergeSchema``.
        pathGlobFilter : str or bool
            an optional glob pattern to only include files with paths matching
            the pattern. The syntax follows `org.apache.hadoop.fs.GlobFilter`.
            It does not change the behavior of
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa
        recursiveFileLookup : str or bool
            recursively scan a directory for files. Using this option
            disables
            `partition discovery <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery>`_.  # noqa

            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedBefore : an optional timestamp to only include files with
            modification times occurring before the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)
        modifiedAfter : an optional timestamp to only include files with
            modification times occurring after the specified time. The provided timestamp
            must be in the following format: YYYY-MM-DDTHH:mm:ss (e.g. 2020-06-01T13:00:00)

        Examples
        --------
        >>> df = spark.read.orc('python/test_support/sql/orc_partitioned')
        >>> df.dtypes
        [('a', 'bigint'), ('b', 'int'), ('c', 'int')]
        """
        self._set_opts(mergeSchema=mergeSchema, pathGlobFilter=pathGlobFilter,
                       modifiedBefore=modifiedBefore, modifiedAfter=modifiedAfter,
                       recursiveFileLookup=recursiveFileLookup)
        if isinstance(path, str):
            path = [path]
        return self._df(self._jreader.orc(_to_seq(self._spark._sc, path)))

    def jdbc(self, url, table, column=None, lowerBound=None, upperBound=None, numPartitions=None,
             predicates=None, properties=None):
        """
        Construct a :class:`DataFrame` representing the database table named ``table``
        accessible via JDBC URL ``url`` and connection ``properties``.

        Partitions of the table will be retrieved in parallel if either ``column`` or
        ``predicates`` is specified. ``lowerBound``, ``upperBound`` and ``numPartitions``
        is needed when ``column`` is specified.

        If both ``column`` and ``predicates`` are specified, ``column`` will be used.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        url : str
            a JDBC URL of the form ``jdbc:subprotocol:subname``
        table : str
            the name of the table
        column : str, optional
            the name of a column of numeric, date, or timestamp type
            that will be used for partitioning;
            if this parameter is specified, then ``numPartitions``, ``lowerBound``
            (inclusive), and ``upperBound`` (exclusive) will form partition strides
            for generated WHERE clause expressions used to split the column
            ``column`` evenly
        lowerBound : str or int, optional
            the minimum value of ``column`` used to decide partition stride
        upperBound : str or int, optional
            the maximum value of ``column`` used to decide partition stride
        numPartitions : int, optional
            the number of partitions
        predicates : list, optional
            a list of expressions suitable for inclusion in WHERE clauses;
            each one defines one partition of the :class:`DataFrame`
        properties : dict, optional
            a dictionary of JDBC database connection arguments. Normally at
            least properties "user" and "password" with their corresponding values.
            For example { 'user' : 'SYSTEM', 'password' : 'mypassword' }

        Notes
        -----
        Don't create too many partitions in parallel on a large cluster;
        otherwise Spark might crash your external database systems.

        Returns
        -------
        :class:`DataFrame`
        """
        if properties is None:
            properties = dict()
        jprop = JavaClass("java.util.Properties", self._spark._sc._gateway._gateway_client)()
        for k in properties:
            jprop.setProperty(k, properties[k])
        if column is not None:
            assert lowerBound is not None, "lowerBound can not be None when ``column`` is specified"
            assert upperBound is not None, "upperBound can not be None when ``column`` is specified"
            assert numPartitions is not None, \
                "numPartitions can not be None when ``column`` is specified"
            return self._df(self._jreader.jdbc(url, table, column, int(lowerBound), int(upperBound),
                                               int(numPartitions), jprop))
        if predicates is not None:
            gateway = self._spark._sc._gateway
            jpredicates = utils.toJArray(gateway, gateway.jvm.java.lang.String, predicates)
            return self._df(self._jreader.jdbc(url, table, jpredicates, jprop))
        return self._df(self._jreader.jdbc(url, table, jprop))


class DataFrameWriter(OptionUtils):
    """
    Interface used to write a :class:`DataFrame` to external storage systems
    (e.g. file systems, key-value stores, etc). Use :attr:`DataFrame.write`
    to access this.

    .. versionadded:: 1.4
    """
    def __init__(self, df):
        self._df = df
        self._spark = df.sql_ctx
        self._jwrite = df._jdf.write()

    def _sq(self, jsq):
        from pyspark.sql.streaming import StreamingQuery
        return StreamingQuery(jsq)

    def mode(self, saveMode):
        """Specifies the behavior when data or table already exists.

        Options include:

        * `append`: Append contents of this :class:`DataFrame` to existing data.
        * `overwrite`: Overwrite existing data.
        * `error` or `errorifexists`: Throw an exception if data already exists.
        * `ignore`: Silently ignore this operation if data already exists.

        .. versionadded:: 1.4.0

        Examples
        --------
        >>> df.write.mode('append').parquet(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        # At the JVM side, the default value of mode is already set to "error".
        # So, if the given saveMode is None, we will not call JVM-side's mode method.
        if saveMode is not None:
            self._jwrite = self._jwrite.mode(saveMode)
        return self

    def format(self, source):
        """Specifies the underlying output data source.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        source : str
            string, name of the data source, e.g. 'json', 'parquet'.

        Examples
        --------
        >>> df.write.format('json').save(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self._jwrite = self._jwrite.format(source)
        return self

    @since(1.5)
    def option(self, key, value):
        """Adds an output option for the underlying data source.

        You can set the following option(s) for writing files:
            * ``timeZone``: sets the string that indicates a time zone ID to be used to format
                timestamps in the JSON/CSV datasources or partition values. The following
                formats of `timeZone` are supported:

                * Region-based zone ID: It should have the form 'area/city', such as \
                  'America/Los_Angeles'.
                * Zone offset: It should be in the format '(+|-)HH:mm', for example '-08:00' or \
                 '+01:00'. Also 'UTC' and 'Z' are supported as aliases of '+00:00'.

                Other short names like 'CST' are not recommended to use because they can be
                ambiguous. If it isn't set, the current value of the SQL config
                ``spark.sql.session.timeZone`` is used by default.
        """
        self._jwrite = self._jwrite.option(key, to_str(value))
        return self

    @since(1.4)
    def options(self, **options):
        """Adds output options for the underlying data source.

        You can set the following option(s) for writing files:
            * ``timeZone``: sets the string that indicates a time zone ID to be used to format
                timestamps in the JSON/CSV datasources or partition values. The following
                formats of `timeZone` are supported:

                * Region-based zone ID: It should have the form 'area/city', such as \
                  'America/Los_Angeles'.
                * Zone offset: It should be in the format '(+|-)HH:mm', for example '-08:00' or \
                 '+01:00'. Also 'UTC' and 'Z' are supported as aliases of '+00:00'.

                Other short names like 'CST' are not recommended to use because they can be
                ambiguous. If it isn't set, the current value of the SQL config
                ``spark.sql.session.timeZone`` is used by default.
        """
        for k in options:
            self._jwrite = self._jwrite.option(k, to_str(options[k]))
        return self

    def partitionBy(self, *cols):
        """Partitions the output by the given columns on the file system.

        If specified, the output is laid out on the file system similar
        to Hive's partitioning scheme.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        cols : str or list
            name of columns

        Examples
        --------
        >>> df.write.partitionBy('year', 'month').parquet(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        if len(cols) == 1 and isinstance(cols[0], (list, tuple)):
            cols = cols[0]
        self._jwrite = self._jwrite.partitionBy(_to_seq(self._spark._sc, cols))
        return self

    def bucketBy(self, numBuckets, col, *cols):
        """Buckets the output by the given columns.If specified,
        the output is laid out on the file system similar to Hive's bucketing scheme.

        .. versionadded:: 2.3.0

        Parameters
        ----------
        numBuckets : int
            the number of buckets to save
        col : str, list or tuple
            a name of a column, or a list of names.
        cols : str
            additional names (optional). If `col` is a list it should be empty.

        Notes
        -----
        Applicable for file-based data sources in combination with
        :py:meth:`DataFrameWriter.saveAsTable`.

        Examples
        --------
        >>> (df.write.format('parquet')  # doctest: +SKIP
        ...     .bucketBy(100, 'year', 'month')
        ...     .mode("overwrite")
        ...     .saveAsTable('bucketed_table'))
        """
        if not isinstance(numBuckets, int):
            raise TypeError("numBuckets should be an int, got {0}.".format(type(numBuckets)))

        if isinstance(col, (list, tuple)):
            if cols:
                raise ValueError("col is a {0} but cols are not empty".format(type(col)))

            col, cols = col[0], col[1:]

        if not all(isinstance(c, str) for c in cols) or not(isinstance(col, str)):
            raise TypeError("all names should be `str`")

        self._jwrite = self._jwrite.bucketBy(numBuckets, col, _to_seq(self._spark._sc, cols))
        return self

    def sortBy(self, col, *cols):
        """Sorts the output in each bucket by the given columns on the file system.

        .. versionadded:: 2.3.0

        Parameters
        ----------
        col : str, tuple or list
            a name of a column, or a list of names.
        cols : str
            additional names (optional). If `col` is a list it should be empty.

        Examples
        --------
        >>> (df.write.format('parquet')  # doctest: +SKIP
        ...     .bucketBy(100, 'year', 'month')
        ...     .sortBy('day')
        ...     .mode("overwrite")
        ...     .saveAsTable('sorted_bucketed_table'))
        """
        if isinstance(col, (list, tuple)):
            if cols:
                raise ValueError("col is a {0} but cols are not empty".format(type(col)))

            col, cols = col[0], col[1:]

        if not all(isinstance(c, str) for c in cols) or not(isinstance(col, str)):
            raise TypeError("all names should be `str`")

        self._jwrite = self._jwrite.sortBy(col, _to_seq(self._spark._sc, cols))
        return self

    def save(self, path=None, format=None, mode=None, partitionBy=None, **options):
        """Saves the contents of the :class:`DataFrame` to a data source.

        The data source is specified by the ``format`` and a set of ``options``.
        If ``format`` is not specified, the default data source configured by
        ``spark.sql.sources.default`` will be used.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : str, optional
            the path in a Hadoop supported file system
        format : str, optional
            the format used to save
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.
        partitionBy : list, optional
            names of partitioning columns
        **options : dict
            all other string options

        Examples
        --------
        >>> df.write.mode("append").save(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self.mode(mode).options(**options)
        if partitionBy is not None:
            self.partitionBy(partitionBy)
        if format is not None:
            self.format(format)
        if path is None:
            self._jwrite.save()
        else:
            self._jwrite.save(path)

    @since(1.4)
    def insertInto(self, tableName, overwrite=None):
        """Inserts the content of the :class:`DataFrame` to the specified table.

        It requires that the schema of the :class:`DataFrame` is the same as the
        schema of the table.

        Optionally overwriting any existing data.
        """
        if overwrite is not None:
            self.mode("overwrite" if overwrite else "append")
        self._jwrite.insertInto(tableName)

    def saveAsTable(self, name, format=None, mode=None, partitionBy=None, **options):
        """Saves the content of the :class:`DataFrame` as the specified table.

        In the case the table already exists, behavior of this function depends on the
        save mode, specified by the `mode` function (default to throwing an exception).
        When `mode` is `Overwrite`, the schema of the :class:`DataFrame` does not need to be
        the same as that of the existing table.

        * `append`: Append contents of this :class:`DataFrame` to existing data.
        * `overwrite`: Overwrite existing data.
        * `error` or `errorifexists`: Throw an exception if data already exists.
        * `ignore`: Silently ignore this operation if data already exists.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        name : str
            the table name
        format : str, optional
            the format used to save
        mode : str, optional
            one of `append`, `overwrite`, `error`, `errorifexists`, `ignore` \
            (default: error)
        partitionBy : str or list
            names of partitioning columns
        **options : dict
            all other string options
        """
        self.mode(mode).options(**options)
        if partitionBy is not None:
            self.partitionBy(partitionBy)
        if format is not None:
            self.format(format)
        self._jwrite.saveAsTable(name)

    def json(self, path, mode=None, compression=None, dateFormat=None, timestampFormat=None,
             lineSep=None, encoding=None, ignoreNullFields=None):
        """Saves the content of the :class:`DataFrame` in JSON format
        (`JSON Lines text format or newline-delimited JSON <http://jsonlines.org/>`_) at the
        specified path.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : str
            the path in any Hadoop supported file system
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.
        compression : str, optional
            compression codec to use when saving to file. This can be one of the
            known case-insensitive shorten names (none, bzip2, gzip, lz4,
            snappy and deflate).
        dateFormat : str, optional
            sets the string that indicates a date format. Custom date formats
            follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to date type. If None is set, it uses the
            default value, ``yyyy-MM-dd``.
        timestampFormat : str, optional
            sets the string that indicates a timestamp format.
            Custom date formats follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to timestamp type. If None is set, it uses the
            default value, ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
        encoding : str, optional
            specifies encoding (charset) of saved json files. If None is set,
            the default UTF-8 charset will be used.
        lineSep : str, optional
            defines the line separator that should be used for writing. If None is
            set, it uses the default value, ``\\n``.
        ignoreNullFields : str or bool, optional
            Whether to ignore null fields when generating JSON objects.
            If None is set, it uses the default value, ``true``.

        Examples
        --------
        >>> df.write.json(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self.mode(mode)
        self._set_opts(
            compression=compression, dateFormat=dateFormat, timestampFormat=timestampFormat,
            lineSep=lineSep, encoding=encoding, ignoreNullFields=ignoreNullFields)
        self._jwrite.json(path)

    def parquet(self, path, mode=None, partitionBy=None, compression=None):
        """Saves the content of the :class:`DataFrame` in Parquet format at the specified path.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : str
            the path in any Hadoop supported file system
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.
        partitionBy : str or list, optional
            names of partitioning columns
        compression : str, optional
            compression codec to use when saving to file. This can be one of the
            known case-insensitive shorten names (none, uncompressed, snappy, gzip,
            lzo, brotli, lz4, and zstd). This will override
            ``spark.sql.parquet.compression.codec``. If None is set, it uses the
            value specified in ``spark.sql.parquet.compression.codec``.

        Examples
        --------
        >>> df.write.parquet(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self.mode(mode)
        if partitionBy is not None:
            self.partitionBy(partitionBy)
        self._set_opts(compression=compression)
        self._jwrite.parquet(path)

    def text(self, path, compression=None, lineSep=None):
        """Saves the content of the DataFrame in a text file at the specified path.
        The text files will be encoded as UTF-8.

        .. versionadded:: 1.6.0

        Parameters
        ----------
        path : str
            the path in any Hadoop supported file system
        compression : str, optional
            compression codec to use when saving to file. This can be one of the
            known case-insensitive shorten names (none, bzip2, gzip, lz4,
            snappy and deflate).
        lineSep : str, optional
            defines the line separator that should be used for writing. If None is
            set, it uses the default value, ``\\n``.

        The DataFrame must have only one column that is of string type.
        Each row becomes a new line in the output file.
        """
        self._set_opts(compression=compression, lineSep=lineSep)
        self._jwrite.text(path)

    def csv(self, path, mode=None, compression=None, sep=None, quote=None, escape=None,
            header=None, nullValue=None, escapeQuotes=None, quoteAll=None, dateFormat=None,
            timestampFormat=None, ignoreLeadingWhiteSpace=None, ignoreTrailingWhiteSpace=None,
            charToEscapeQuoteEscaping=None, encoding=None, emptyValue=None, lineSep=None):
        r"""Saves the content of the :class:`DataFrame` in CSV format at the specified path.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        path : str
            the path in any Hadoop supported file system
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.

        compression : str, optional
            compression codec to use when saving to file. This can be one of the
            known case-insensitive shorten names (none, bzip2, gzip, lz4,
            snappy and deflate).
        sep : str, optional
            sets a separator (one or more characters) for each field and value. If None is
            set, it uses the default value, ``,``.
        quote : str, optional
            sets a single character used for escaping quoted values where the
            separator can be part of the value. If None is set, it uses the default
            value, ``"``. If an empty string is set, it uses ``u0000`` (null character).
        escape : str, optional
            sets a single character used for escaping quotes inside an already
            quoted value. If None is set, it uses the default value, ``\``
        escapeQuotes : str or bool, optional
            a flag indicating whether values containing quotes should always
            be enclosed in quotes. If None is set, it uses the default value
            ``true``, escaping all values containing a quote character.
        quoteAll : str or bool, optional
            a flag indicating whether all values should always be enclosed in
            quotes. If None is set, it uses the default value ``false``,
            only escaping values containing a quote character.
        header : str or bool, optional
            writes the names of columns as the first line. If None is set, it uses
            the default value, ``false``.
        nullValue : str, optional
            sets the string representation of a null value. If None is set, it uses
            the default value, empty string.
        dateFormat : str, optional
            sets the string that indicates a date format. Custom date formats follow
            the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to date type. If None is set, it uses the
            default value, ``yyyy-MM-dd``.
        timestampFormat : str, optional
            sets the string that indicates a timestamp format.
            Custom date formats follow the formats at
            `datetime pattern <https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html>`_.  # noqa
            This applies to timestamp type. If None is set, it uses the
            default value, ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
        ignoreLeadingWhiteSpace : str or bool, optional
            a flag indicating whether or not leading whitespaces from
            values being written should be skipped. If None is set, it
            uses the default value, ``true``.
        ignoreTrailingWhiteSpace : str or bool, optional
            a flag indicating whether or not trailing whitespaces from
            values being written should be skipped. If None is set, it
            uses the default value, ``true``.
        charToEscapeQuoteEscaping : str, optional
            sets a single character used for escaping the escape for
            the quote character. If None is set, the default value is
            escape character when escape and quote characters are
            different, ``\0`` otherwise..
        encoding : str, optional
            sets the encoding (charset) of saved csv files. If None is set,
            the default UTF-8 charset will be used.
        emptyValue : str, optional
            sets the string representation of an empty value. If None is set, it uses
            the default value, ``""``.
        lineSep : str, optional
            defines the line separator that should be used for writing. If None is
            set, it uses the default value, ``\\n``. Maximum length is 1 character.

        Examples
        --------
        >>> df.write.csv(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self.mode(mode)
        self._set_opts(compression=compression, sep=sep, quote=quote, escape=escape, header=header,
                       nullValue=nullValue, escapeQuotes=escapeQuotes, quoteAll=quoteAll,
                       dateFormat=dateFormat, timestampFormat=timestampFormat,
                       ignoreLeadingWhiteSpace=ignoreLeadingWhiteSpace,
                       ignoreTrailingWhiteSpace=ignoreTrailingWhiteSpace,
                       charToEscapeQuoteEscaping=charToEscapeQuoteEscaping,
                       encoding=encoding, emptyValue=emptyValue, lineSep=lineSep)
        self._jwrite.csv(path)

    def orc(self, path, mode=None, partitionBy=None, compression=None):
        """Saves the content of the :class:`DataFrame` in ORC format at the specified path.

        .. versionadded:: 1.5.0

        Parameters
        ----------
        path : str
            the path in any Hadoop supported file system
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.
        partitionBy : str or list, optional
            names of partitioning columns
        compression : str, optional
            compression codec to use when saving to file. This can be one of the
            known case-insensitive shorten names (none, snappy, zlib, and lzo).
            This will override ``orc.compress`` and
            ``spark.sql.orc.compression.codec``. If None is set, it uses the value
            specified in ``spark.sql.orc.compression.codec``.

        Examples
        --------
        >>> orc_df = spark.read.orc('python/test_support/sql/orc_partitioned')
        >>> orc_df.write.orc(os.path.join(tempfile.mkdtemp(), 'data'))
        """
        self.mode(mode)
        if partitionBy is not None:
            self.partitionBy(partitionBy)
        self._set_opts(compression=compression)
        self._jwrite.orc(path)

    def jdbc(self, url, table, mode=None, properties=None):
        """Saves the content of the :class:`DataFrame` to an external database table via JDBC.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        url : str
            a JDBC URL of the form ``jdbc:subprotocol:subname``
        table : str
            Name of the table in the external database.
        mode : str, optional
            specifies the behavior of the save operation when data already exists.

            * ``append``: Append contents of this :class:`DataFrame` to existing data.
            * ``overwrite``: Overwrite existing data.
            * ``ignore``: Silently ignore this operation if data already exists.
            * ``error`` or ``errorifexists`` (default case): Throw an exception if data already \
                exists.
        properties : dict
            a dictionary of JDBC database connection arguments. Normally at
            least properties "user" and "password" with their corresponding values.
            For example { 'user' : 'SYSTEM', 'password' : 'mypassword' }

        Notes
        -----
        Don't create too many partitions in parallel on a large cluster;
        otherwise Spark might crash your external database systems.
        """
        if properties is None:
            properties = dict()
        jprop = JavaClass("java.util.Properties", self._spark._sc._gateway._gateway_client)()
        for k in properties:
            jprop.setProperty(k, properties[k])
        self.mode(mode)._jwrite.jdbc(url, table, jprop)


class DataFrameWriterV2(object):
    """
    Interface used to write a class:`pyspark.sql.dataframe.DataFrame`
    to external storage using the v2 API.

    .. versionadded:: 3.1.0
    """

    def __init__(self, df, table):
        self._df = df
        self._spark = df.sql_ctx
        self._jwriter = df._jdf.writeTo(table)

    @since(3.1)
    def using(self, provider):
        """
        Specifies a provider for the underlying output data source.
        Spark's default catalog supports "parquet", "json", etc.
        """
        self._jwriter.using(provider)
        return self

    @since(3.1)
    def option(self, key, value):
        """
        Add a write option.
        """
        self._jwriter.option(key, to_str(value))
        return self

    @since(3.1)
    def options(self, **options):
        """
        Add write options.
        """
        options = {k: to_str(v) for k, v in options.items()}
        self._jwriter.options(options)
        return self

    @since(3.1)
    def tableProperty(self, property, value):
        """
        Add table property.
        """
        self._jwriter.tableProperty(property, value)
        return self

    @since(3.1)
    def partitionedBy(self, col, *cols):
        """
        Partition the output table created by `create`, `createOrReplace`, or `replace` using
        the given columns or transforms.

        When specified, the table data will be stored by these values for efficient reads.

        For example, when a table is partitioned by day, it may be stored
        in a directory layout like:

        * `table/day=2019-06-01/`
        * `table/day=2019-06-02/`

        Partitioning is one of the most widely used techniques to optimize physical data layout.
        It provides a coarse-grained index for skipping unnecessary data reads when queries have
        predicates on the partitioned columns. In order for partitioning to work well, the number
        of distinct values in each column should typically be less than tens of thousands.

        `col` and `cols` support only the following functions:

        * :py:func:`pyspark.sql.functions.years`
        * :py:func:`pyspark.sql.functions.months`
        * :py:func:`pyspark.sql.functions.days`
        * :py:func:`pyspark.sql.functions.hours`
        * :py:func:`pyspark.sql.functions.bucket`

        """
        col = _to_java_column(col)
        cols = _to_seq(self._spark._sc, [_to_java_column(c) for c in cols])
        return self

    @since(3.1)
    def create(self):
        """
        Create a new table from the contents of the data frame.

        The new table's schema, partition layout, properties, and other configuration will be
        based on the configuration set on this writer.
        """
        self._jwriter.create()

    @since(3.1)
    def replace(self):
        """
        Replace an existing table with the contents of the data frame.

        The existing table's schema, partition layout, properties, and other configuration will be
        replaced with the contents of the data frame and the configuration set on this writer.
        """
        self._jwriter.replace()

    @since(3.1)
    def createOrReplace(self):
        """
        Create a new table or replace an existing table with the contents of the data frame.

        The output table's schema, partition layout, properties,
        and other configuration will be based on the contents of the data frame
        and the configuration set on this writer.
        If the table exists, its configuration and data will be replaced.
        """
        self._jwriter.createOrReplace()

    @since(3.1)
    def append(self):
        """
        Append the contents of the data frame to the output table.
        """
        self._jwriter.append()

    @since(3.1)
    def overwrite(self, condition):
        """
        Overwrite rows matching the given filter condition with the contents of the data frame in
        the output table.
        """
        self._jwriter.overwrite(condition)

    @since(3.1)
    def overwritePartitions(self):
        """
        Overwrite all partition for which the data frame contains at least one row with the contents
        of the data frame in the output table.

        This operation is equivalent to Hive's `INSERT OVERWRITE ... PARTITION`, which replaces
        partitions dynamically depending on the contents of the data frame.
        """
        self._jwriter.overwritePartitions()


def _test():
    import doctest
    import os
    import tempfile
    import py4j
    from pyspark.context import SparkContext
    from pyspark.sql import SparkSession
    import pyspark.sql.readwriter

    os.chdir(os.environ["SPARK_HOME"])

    globs = pyspark.sql.readwriter.__dict__.copy()
    sc = SparkContext('local[4]', 'PythonTest')
    try:
        spark = SparkSession.builder.getOrCreate()
    except py4j.protocol.Py4JError:
        spark = SparkSession(sc)

    globs['tempfile'] = tempfile
    globs['os'] = os
    globs['sc'] = sc
    globs['spark'] = spark
    globs['df'] = spark.read.parquet('python/test_support/sql/parquet_partitioned')
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.readwriter, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF)
    sc.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



import logging

from pyspark.sql import SparkSession

from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark.offline_store import DeltaMetadataWriter


logger = logging.getLogger(__name__)


def run_delta_maintenance(spark: SparkSession, materialization_task_params: MaterializationTaskParams):
    path = materialization_task_params.offline_store_path
    delta_metadata_writer = DeltaMetadataWriter(spark)
    delta_maintenance_params = materialization_task_params.delta_maintenance_task_info.delta_maintenance_parameters
    if delta_maintenance_params.generate_manifest:
        # regardless, the manual run is necessary because the auto generation
        # is not safe to race conditions. see https://docs.delta.io/latest/presto-integration.html#language-python
        delta_metadata_writer.generate_symlink_manifest(path)
    if delta_maintenance_params.execute_compaction:
        delta_metadata_writer.optimize_execute_compaction(path)
    if delta_maintenance_params.vacuum:
        delta_metadata_writer.vacuum(path)



import json
import logging
import re
import subprocess
import sys
from pathlib import Path
from typing import List

from tecton_core.errors import FailedDependencyDownloadError


PLATFORM = "linux_x86_64"

# need to list out compatible manylinux wheel for pip (https://github.com/pypa/pip/issues/10760)
COMPATIBLE_PLATFORMS = ["linux_x86_64", "manylinux1_x86_64", "manylinux_2_17_x86_64", "manylinux2014_x86_64"]

# additional wheel distributions for libs that do not have wheels in pypi
ADDITIONAL_WHEELS_REPOS = [
    "https://s3.us-west-2.amazonaws.com/tecton.ai.public/python/PyPika-0.48.9-py2.py3-none-any.whl"
]

PYTHON_VERSION_TO_PLATFORM = {
    "3.8": f"{PLATFORM}-cp-3.8.17-cp38",
    "3.9": f"{PLATFORM}-cp-3.9.17-cp39",
}

MISSING_REQUIREMENTS_ERROR = "Could not find a version that satisfies the requirement"
ENSURE_WHEELS_EXIST_WARNING = "Please also ensure that the package(s) have wheels (.whl) available for download in PyPI or any other repository used."

logger = logging.getLogger(__name__)


def resolve_dependencies(
    requirements_path: Path,
    lock_output_path: Path,
    resolved_requirements_path: Path,
    python_version: str,
    timeout_seconds: int,
):
    """Resolve dependencies using `pex`
    Parameters:
        requirements_path(Path): Path to the `requirements.txt` file
        lock_output_path(Path): Path to store the output lock.json
        resolved_requirements_path(Path): The target path for generating the fully resolved and pinned `resolved-requirements.txt` file
        python_version(str): The python version to resolve dependencies for
        timeout_seconds(int): The timeout in seconds for the dependency resolution
    """
    major_minor_version = _get_major_minor_version(python_version)
    if major_minor_version not in PYTHON_VERSION_TO_PLATFORM:
        msg = f"Invalid `python_version` {python_version}. Expected one of: {list(PYTHON_VERSION_TO_PLATFORM.keys())}"
        raise ValueError(msg)
    platform = PYTHON_VERSION_TO_PLATFORM[major_minor_version]
    logger.debug(f"Resolving dependencies for platform: {platform}")
    lock_command = _construct_lock_command(
        requirements_path=requirements_path, target_path=lock_output_path, platform=platform
    )
    export_command = _construct_export_command(
        target_path=lock_output_path,
        resolved_requirements_path=resolved_requirements_path,
        platform=platform,
    )
    _run_pex_command(command_list=lock_command, timeout_seconds=timeout_seconds)
    _run_pex_command(command_list=export_command, timeout_seconds=timeout_seconds)


def download_dependencies(requirements_path: Path, target_directory: Path, python_version: str):
    """
    Download wheels for all dependencies in a requirements.txt to a target directory
    Parameters:
        requirements_path(Path): Path to requirements.txt
        target_directory(Path): The target directory to download requirements to
        python_version(str): The python version to download dependencies for
    """
    command = _construct_download_command(
        target_path=target_directory, requirements_path=requirements_path, python_version=python_version
    )

    logger.debug(f"Executing command: {' '.join(command)}")

    result = subprocess.run(
        command,
        text=True,
    )
    if result.returncode != 0:
        raise FailedDependencyDownloadError(result.stderr)


def create_requirements_from_lock_file(lock_file, requirements_path):
    lock_data = json.loads(lock_file.read_text())
    with open(requirements_path, "w") as requirements_file:
        for resolve in lock_data["locked_resolves"]:
            for requirement in resolve["locked_requirements"]:
                for artifact in requirement["artifacts"]:
                    url = artifact["url"]
                    hash_value = f"{artifact['algorithm']}:{artifact['hash']}"
                    requirements_file.write(f"{url} --hash={hash_value}\n")


def _get_major_minor_version(version: str):
    version_parts = version.split(".")
    return ".".join(version_parts[:2])


def _run_pex_command(command_list: List[str], timeout_seconds: int):
    """Run the `pex` command passed as input and process any errors
    Parameters:
        command_list(str): The pex command to be executed
        timeout_seconds(int): The timeout in seconds for the pex command
    """
    command = [sys.executable, "-m", "tecton.cli.pex_wrapper", *command_list]
    logger.debug(f"Executing command: {' '.join(command)}")
    try:
        result = subprocess.run(
            command,
            capture_output=True,
            text=True,
            timeout=timeout_seconds,
        )
    except subprocess.TimeoutExpired:
        error_message = (
            "Dependency Resolution timed out! If problem persists, please contact Tecton Support for assistance"
        )
        raise TimeoutError(error_message)
    if result.returncode != 0:
        logger.debug(f"Raw pex error: {result.stderr}")
        cleaned_error = _parse_pex_error(result.stderr)
        raise ValueError(cleaned_error)


def _construct_lock_command(requirements_path: Path, target_path: Path, platform: str) -> List[str]:
    return [
        "lock",
        "create",
        "-r",
        str(requirements_path),
        "--no-build",
        "--style=strict",
        "--resolver-version=pip-2020-resolver",
        "-o",
        str(target_path),
        "--platform",
        platform,
    ] + [item for repo in ADDITIONAL_WHEELS_REPOS for item in ("-f", repo)]


def _construct_export_command(target_path: Path, resolved_requirements_path: Path, platform: str) -> List[str]:
    return ["lock", "export", "--platform", platform, str(target_path), "--output", str(resolved_requirements_path)]


def _construct_download_command(target_path: Path, requirements_path: Path, python_version: str):
    return [
        sys.executable,
        "-m",
        "pip",
        "download",
        "-r",
        str(requirements_path),
        "-d",
        str(target_path),
        "--no-cache-dir",
        "--only-binary",
        ":all:",
        "--python-version",
        python_version,
        "--implementation",
        "cp",
    ] + [item for platform in COMPATIBLE_PLATFORMS for item in ("--platform", platform)]


def _parse_pex_error(error_string: str) -> str:
    """Parse and cleanup error messages from the `pex` command"""
    start_index = error_string.find("ERROR:")
    if start_index != -1:
        error_string = error_string[start_index + 6 :].replace("\n", " ")
    # The pex error message does not clarify that wheels must be present and so we append it to the original error message
    if MISSING_REQUIREMENTS_ERROR in error_string:
        error_string = f"{error_string}\n\n💡 {ENSURE_WHEELS_EXIST_WARNING}"
    return error_string


def is_valid_environment_name(name: str) -> bool:
    # Only letters, numbers, hyphens, or underscores allowed in an environment name
    pattern = r"^[a-zA-Z0-9_-]+$"
    return bool(re.match(pattern, name))


class EnvironmentDependencies:
    def __init__(self, data):
        self.data = data
        self.locked_dependencies = {
            item["project_name"]: item for item in data["locked_resolves"][0]["locked_requirements"]
        }
        self.input_requirements = data["requirements"]

    def get_version(self, dependency_name):
        """Retrieve the version of the specified dependency"""
        package = self.locked_dependencies.get(dependency_name)
        if package:
            return package["version"]
        return None

    def is_dependency_present(self, dependency_name):
        """Check if the specified dependency is present in the requirements."""
        return dependency_name in self.locked_dependencies

    def get_dependency_extras(self, dependency_name):
        """Retrieve a list of extras available for the specified dependency."""
        package = self.locked_dependencies.get(dependency_name)
        if package:
            extras = set()
            for dist in package["requires_dists"]:
                if "; extra ==" in dist:
                    extracted_extra = dist.split("; extra ==")[1].strip(' "').replace("'", "")
                    extras.add(extracted_extra)
            return list(extras)
        return []

    def is_version_pinned(self, package_name):
        """Check if the dependency version is explicitly pinned in the input requirements"""
        for req in self.input_requirements:
            if package_name in req and "==" in req:
                return True
        return False



import re
import sys
import typing
from dataclasses import dataclass
from textwrap import dedent
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pandas
import pendulum

from tecton_core import specs
from tecton_core.data_types import ArrayType
from tecton_core.data_types import BoolType
from tecton_core.data_types import DataType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.errors import UDF_ERROR
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonSnowflakeNotImplementedError
from tecton_core.id_helper import IdHelper
from tecton_core.materialization_context import MaterializationContext
from tecton_core.pipeline_common import CONSTANT_TYPE
from tecton_core.pipeline_common import CONSTANT_TYPE_OBJECTS
from tecton_core.pipeline_common import check_transformation_type
from tecton_core.pipeline_common import constant_node_to_value
from tecton_core.pipeline_common import get_keyword_inputs
from tecton_core.pipeline_common import get_time_window_from_data_source_node
from tecton_core.pipeline_common import positional_inputs
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationArgs
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_proto.common.data_source_type__client_pb2 import DataSourceType
from tecton_proto.data.transformation__client_pb2 import Transformation
from tecton_snowflake.templates_utils import load_template
from tecton_snowflake.utils import format_sql
from tecton_snowflake.utils import generate_random_name


if typing.TYPE_CHECKING:
    import snowflake.snowpark


PIPELINE_TEMPLATE = load_template("transformation_pipeline.sql")
TIME_LIMIT_TEMPLATE = load_template("time_limit.sql")
DATA_SOURCE_TEMPLATE = load_template("data_source.sql")
TEMP_CTE_PREFIX = "_TT_CTE_"
TEMP_DS_PREFIX = "_TT_DS_"
SPINE_TABLE_NAME = "_TT_SPINE_TABLE"
UDF_OUTPUT_COLUMN_NAME = "_TT_UDF_OUTPUT"
TEMP_PIPELINE_VIEW_NAME = "_TEMP_PIPELINE_VIEW_FROM_DF"
TRANSFORMATION_INPUT_PARAMETER = "transformation_input"
TRANSFORMATION_COLUMN_PREFIX = "_UDF_INTERNAL_TRANSFORMATION__"

SEPERATOR = "__"

SPARK_TO_SNOWFLAKE_TYPES = {
    Int64Type(): "NUMBER",
    Float64Type(): "FLOAT",
    StringType(): "STRING",
    BoolType(): "BOOLEAN",
    ArrayType(Int64Type()): "ARRAY",
    ArrayType(Float32Type()): "ARRAY",
    ArrayType(Float64Type()): "ARRAY",
    ArrayType(StringType()): "ARRAY",
}


# TODO(deprecate_after=0.7): Snowpark mode transformations are not supported in versions <=0.8. We should switch to only using PipelineSqlBuilder in the future
def pipeline_to_sql_string(
    pipeline: Pipeline,
    data_sources: List[specs.DataSourceSpec],
    transformations: List[specs.TransformationSpec],
    materialization_context: MaterializationContext,
    mock_sql_inputs: Optional[Dict[str, str]] = None,
    session: "snowflake.snowpark.Session" = None,
) -> str:
    """This method is deprecated!

    Do not use if not necessary. Instead, use PipelineSqlBuilder."""
    return format_sql(
        _PipelineBuilder(
            pipeline=pipeline,
            data_sources=data_sources,
            transformations=transformations,
            mock_sql_inputs=mock_sql_inputs,
            materialization_context=materialization_context,
            session=session,
        ).get_sql_string()
    )


# Pandas Pipeline (ODFV)
# input_df (snowpark df) is the spine passed in by the user (including request context),
# and it has been augmented with dependent fv fields in of the form "_udf_internal_{input_name}.{feature_field_name}".
# The dataframe we return will be everything from the spine, with the on-demand features added
# TODO: Figure out a way to get the dependent fv fields on tecton apply, this way we can
#      register the udf on apply and avoid doing it adhoc on feature retrival.
def pipeline_to_df_with_input(
    session: "snowflake.snowpark.Session",
    # This should have data from all inputs
    input_df: "snowflake.snowpark.DataFrame",
    pipeline: Pipeline,
    transformations: List[specs.TransformationSpec],
    output_schema: Dict[str, DataType],
    name: str,
    fv_id: str,
    namespace: str,
    append_prefix: bool,
) -> "snowflake.snowpark.DataFrame":
    # TODO: Currently there's a bug in toPandas() call, types may not be casted to the correct type.
    # e.g. Long is currently being casted to object(decimal.Decimal) instead of int64.
    return _ODFVPipelineBuilder(
        session=session,
        input_df=input_df,
        output_schema=output_schema,
        name=name,
        namespace=namespace,
        pipeline=pipeline,
        transformations=transformations,
        fv_id=fv_id,
        append_prefix=append_prefix,
    ).get_df()


# Used for j2 template
@dataclass
class _NodeInput:
    name: str
    sql_str: str


def has_snowpark_transformation(transformations: List[specs.TransformationSpec]) -> bool:
    for transformation in transformations:
        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SNOWPARK:
            return True
    return False


def has_push_source(data_sources: List[specs.DataSourceSpec]) -> bool:
    for data_source in data_sources:
        if data_source.type in (DataSourceType.PUSH_NO_BATCH, DataSourceType.PUSH_WITH_BATCH):
            return True
    return False


# TODO(deprecate_after=0.7): Snowpark mode transformations are not supported in versions <=0.8. We should switch to only using PipelineSqlBuilder in the future
class _PipelineBuilder:
    """
    This class is for Snowflake Pipelines with snowpark transformations.

    WARNING: This class is deprecated! PipelineSqlBuilder should be used instead.
    """

    # The value of internal nodes in the tree
    _VALUE_TYPE = Union[str, CONSTANT_TYPE, MaterializationContext, "snowflake.snowpark.DataFrame"]

    def __init__(
        self,
        pipeline: Pipeline,
        data_sources: List[specs.DataSourceSpec],
        # we only use mode and name from these
        transformations: Union[List[Transformation], List[TransformationArgs]],
        mock_sql_inputs: Optional[Dict[str, str]],
        materialization_context: MaterializationContext,
        session: Optional["snowflake.snowpark.Session"],
    ):
        self._pipeline = pipeline
        self._id_to_ds = {ds.id: ds for ds in data_sources}
        self._id_to_transformation = {t.id: t for t in transformations}
        self._mock_sql_inputs = mock_sql_inputs
        self._materialization_context = materialization_context
        self._has_snowpark = has_snowpark_transformation(transformations)
        self._session = session
        self._ds_to_sql_str = {}
        if self._has_snowpark:
            assert self._session is not None

    def get_sql_string(self) -> str:
        if self._has_snowpark:
            df = self._node_to_value(self._pipeline.root)
            temp_pipeline_view_name = generate_random_name()
            df.create_or_replace_temp_view(temp_pipeline_view_name)
            row = self._session.sql(f"SELECT GET_DDL('VIEW', '{temp_pipeline_view_name}') AS SQL").collect(
                statement_params={"SF_PARTNER": "tecton-ai"}
            )[0]
            generated_sql = row.as_dict()["SQL"]
            # The generated sql query will be something like:
            #
            # create or replace view TEST_VIEW(
            # 	USER_ID,
            # 	USER_HAS_GOOD_CREDIT,
            # 	TIMESTAMP
            # ) as  SELECT  *  FROM (
            #   xxx
            #   xxx
            #   xxx);
            # Remove the "create or replace view XXX as" part
            m = re.search(f"create or replace view {temp_pipeline_view_name}\([^()]*\) as ((?s:.)*);", generated_sql)
            if m:
                view_sql = m.group(1)
            else:
                msg = f"Couldn't extract from generated sql query: {generated_sql}"
                raise TectonInternalError(msg)
            self._session.sql(f"DROP VIEW IF EXISTS {temp_pipeline_view_name}")
            return view_sql
        else:
            sql_str = self._node_to_value(self._pipeline.root)
            assert isinstance(sql_str, str)
            sql_str = DATA_SOURCE_TEMPLATE.render(data_sources=self._ds_to_sql_str, source=sql_str)
            return sql_str

    def _node_to_value(self, pipeline_node: PipelineNode) -> _VALUE_TYPE:
        if pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_value(pipeline_node.transformation_node)
        elif pipeline_node.HasField("data_source_node"):
            return self._data_source_node_to_value(pipeline_node.data_source_node)
        elif pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("materialization_context_node"):
            return self._materialization_context
        elif pipeline_node.HasField("request_data_source_node"):
            msg = "RequestDataSource is not supported in Snowflake SQL pipelines"
            raise TectonSnowflakeNotImplementedError(msg)
        elif pipeline_node.HasField("feature_view_node"):
            msg = "Dependent FeatureViews are not supported in Snowflake SQL pipelines"
            raise TectonSnowflakeNotImplementedError(msg)
        else:
            msg = f"Unknown PipelineNode type: {pipeline_node}"
            raise KeyError(msg)

    def _data_source_node_to_value(
        self, data_source_node: DataSourceNode
    ) -> Union[str, "snowflake.snowpark.DataFrame"]:
        """Creates a sql string from a ds and time parameters."""
        sql_str = ""
        if self._mock_sql_inputs is not None and data_source_node.input_name in self._mock_sql_inputs:
            sql_str = self._mock_sql_inputs[data_source_node.input_name]
        else:
            time_window = get_time_window_from_data_source_node(
                feature_time_limits=pendulum.Period(
                    self._materialization_context.start_time, self._materialization_context.end_time
                ),
                schedule_interval=self._materialization_context.batch_schedule,
                data_source_node=data_source_node,
            )
            ds = self._id_to_ds[IdHelper.to_string(data_source_node.virtual_data_source_id)]
            sql_str = self._get_ds_sql_str(ds, time_window)

        cte_name = f"{TEMP_DS_PREFIX}_{generate_random_name()}"
        self._ds_to_sql_str[cte_name] = sql_str
        if self._has_snowpark:
            return self._session.sql(sql_str)
        else:
            return cte_name

    def _get_ds_sql_str(self, ds: specs.DataSourceSpec, time_window: Optional[pendulum.Period]) -> str:
        if not ds.batch_source:
            msg = "Snowflake SQL pipeline only supports batch data source"
            raise TectonSnowflakeNotImplementedError(msg)
        if not isinstance(ds.batch_source, specs.SnowflakeSourceSpec):
            msg = f"Snowflake SQL pipeline does not support batch data source: {ds.batch_source}"
            raise TectonSnowflakeNotImplementedError(msg)

        snowflake_source = ds.batch_source

        if snowflake_source.table:
            if snowflake_source.query:
                msg = "Only one of table and query can be specified"
                raise TectonInternalError(msg)
            # Makes sure we have all the info for the table
            assert snowflake_source.database
            assert snowflake_source.schema
            sql_str = f"{snowflake_source.database}.{snowflake_source.schema}.{snowflake_source.table}"
        elif snowflake_source.query:
            sql_str = snowflake_source.query
        else:
            msg = "Either table or query must be specified for Snowflake data source"
            raise TectonInternalError(msg)

        if time_window is None:
            return f"SELECT * FROM ({sql_str})"

        # If we have a time window, we need to filter the source based on it
        if not snowflake_source.timestamp_field:
            msg = f"timestamp_field must be set within Snowflake data source '{ds.name}' to use time filtering in this feature view."
            raise TectonInternalError(msg)
        return TIME_LIMIT_TEMPLATE.render(
            source=sql_str,
            timestamp_key=snowflake_source.timestamp_field,
            start_time=time_window.start,
            end_time=time_window.end,
        )

    def _transformation_node_to_value(
        self, transformation_node: TransformationNode
    ) -> Union[str, "snowflake.snowpark.DataFrame"]:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(
        self, transformation_node, args, kwargs
    ) -> Union[str, "snowflake.snowpark.DataFrame"]:
        """For the given transformation node, returns the corresponding sql string or dataframe."""
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = transformation.user_function

        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL:
            return self._wrap_sql_function(transformation_node, user_function)(*args, **kwargs)
        elif transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SNOWPARK:
            res = user_function(*args, **kwargs)
            check_transformation_type(transformation.name, res, "snowpark", self._possible_modes())
            return res
        else:
            msg = f"Invalid transformation mode: {TransformationMode.Name(transformation.transformation_mode)} for a Snowflake SQL pipeline"
            raise KeyError(msg)

    def _wrap_sql_function(
        self, transformation_node: TransformationNode, user_function: Callable[..., str]
    ) -> Callable[..., str]:
        def wrapped(*args, **kwargs):
            transformationInputs = []
            wrapped_args = []
            for arg, node_input in zip(args, positional_inputs(transformation_node)):
                input_str, is_sql = self._wrap_node_inputvalue(node_input, arg)
                cte_name = TEMP_CTE_PREFIX + generate_random_name()
                if is_sql:
                    wrapped_args.append(cte_name)
                    transformationInputs.append(_NodeInput(name=cte_name, sql_str=input_str))
                else:
                    wrapped_args.append(input_str)
            keyword_inputs = get_keyword_inputs(transformation_node)
            wrapped_kwargs = {}
            for k, v in kwargs.items():
                node_input = keyword_inputs[k]
                input_str, is_sql = self._wrap_node_inputvalue(node_input, v)
                if is_sql:
                    cte_name = TEMP_CTE_PREFIX + k
                    wrapped_kwargs[k] = cte_name
                    transformationInputs.append(_NodeInput(name=cte_name, sql_str=input_str))
                else:
                    wrapped_kwargs[k] = input_str
            user_function_sql = dedent(user_function(*wrapped_args, **wrapped_kwargs))
            sql_str = PIPELINE_TEMPLATE.render(inputs=transformationInputs, user_function=user_function_sql)
            transformation_name = self._id_to_transformation[
                IdHelper.to_string(transformation_node.transformation_id)
            ].name
            check_transformation_type(transformation_name, sql_str, "snowflake_sql", self._possible_modes())
            return sql_str

        return wrapped

    def _wrap_node_inputvalue(self, node_input, value: _VALUE_TYPE) -> Tuple[Union[CONSTANT_TYPE], bool]:
        """Returns the node value, along with a boolean indicating whether the input is a sql str."""
        if node_input.node.HasField("constant_node"):
            assert value is None or isinstance(value, CONSTANT_TYPE_OBJECTS)
            return value, False
        elif node_input.node.HasField("data_source_node"):
            # For data source we don't want a bracket around it
            assert isinstance(value, str)
            return value, False
        elif node_input.node.HasField("materialization_context_node"):
            assert isinstance(value, MaterializationContext)
            return value, False
        else:
            # This should be a sql string already, we need to return this with a bracket wrapping it
            # The current implementation will add a round bracket () to all subquery
            assert isinstance(value, str)
            return f"({value})", True

    def _possible_modes(self):
        return ["snowflake_sql", "pipeline", "snowpark"]


# This class is for Pandas pipelines
class _ODFVPipelineBuilder:
    def __init__(
        self,
        session: "snowflake.snowpark.Session",
        name: str,
        namespace: str,
        transformations: List[specs.TransformationSpec],
        output_schema: Dict[str, DataType],
        pipeline: Pipeline,
        fv_id: str,
        input_df: "snowflake.snowpark.DataFrame" = None,
        append_prefix: bool = True,
    ):
        self._input_df = input_df
        self._session = session
        self._pipeline = pipeline
        self._name = name
        self._namespace = namespace
        self._fv_id = fv_id
        self._id_to_transformation = {t.id: t for t in transformations}
        self._output_schema = output_schema
        self._append_prefix = append_prefix

    def get_df(self) -> "snowflake.snowpark.DataFrame":
        if not self._pipeline.root.HasField("transformation_node"):
            msg = "Root pipeline has to be a transformation for pandas mode"
            raise ValueError(msg)
        output_df, _ = self._transformation_node_to_df(
            self._pipeline.root.transformation_node, is_top_level_transformation=True
        )
        return output_df

    def _transformation_node_to_df(
        self, transformation_node: TransformationNode, is_top_level_transformation: bool = False
    ) -> ("snowflake.snowpark.DataFrame", list):
        """
        Run pipeline on transformation node
        :param transformation_node: input transformation node
        :param is_top_level_transformation: whether the transformation node is the root node of the pipeline
        :return: 1. A snowpark DataFrame with the results from the transformation
                2. The list of the column names of this DataFrame.
                (We return them specifically since Snowflake will automatically capitalize them in a way the code does not expect.
        """
        # Columns in snowflake dataframe have double quotes around them.
        udf_args = [c.strip('"') for c in self._input_df.columns if ("_UDF_INTERNAL" in c)]

        # The following parameters are used to generate the UDF snowflake calls.
        input_columns = []  # Columns on the input dataframe that are used as inputs to the transformation.
        input_map = {}  # Maps parameter names to relevant featuere columns/keys.
        prefix_map = {}  # Maps parameter names to the prefix used on their dependent features.
        input_df = self._input_df  # The DataFrame to be passed to the UDF

        # Input for On-Demand can only be a feature view, request data source, or transformation
        for transformation_input in transformation_node.inputs:
            input_node = transformation_input.node
            if input_node.HasField("feature_view_node"):
                features = []
                feature_view_node = input_node.feature_view_node
                prefix = f"_UDF_INTERNAL_{feature_view_node.input_name}_{self._fv_id}__".upper()
                for feature in udf_args:
                    if not feature.startswith(prefix):
                        continue
                    input_columns.append(feature)
                    features.append(feature)
                input_map[feature_view_node.input_name] = features
                prefix_map[feature_view_node.input_name] = prefix
            elif input_node.HasField("request_data_source_node"):
                request_data_source_node = input_node.request_data_source_node
                field_names = [field.name for field in request_data_source_node.request_context.tecton_schema.columns]
                for input_col in field_names:
                    input_columns.append(input_col)
                input_map[request_data_source_node.input_name] = field_names
                prefix_map[request_data_source_node.input_name] = ""
                # Request context should be in the input_df already
            elif input_node.HasField("transformation_node"):
                # TODO: https://tecton.atlassian.net/browse/TEC-11231
                assert (
                    len(transformation_node.inputs) == 1
                ), "Snowflake currently only supports a single transformation as the input to another transformation"
                # overwrite the input df to be the output of the nested transformation
                input_df, result_columns = self._transformation_node_to_df(input_node.transformation_node)
                prefix = TRANSFORMATION_COLUMN_PREFIX
                features = []
                for feature in result_columns:
                    if not feature.startswith(prefix):
                        continue
                    input_columns.append(feature)
                    features.append(feature)
                # We don't know the param name for nested transformations so we set a placeholder name.
                input_map[TRANSFORMATION_INPUT_PARAMETER] = features
                prefix_map[TRANSFORMATION_INPUT_PARAMETER] = prefix
            else:
                msg = "Snowflake only supports feature view and request data source as input."
                raise TectonSnowflakeNotImplementedError(msg)
        # Get back the name of the UDF
        ondemand_udf = self._generate_on_demand_udf(transformation_node, input_map, prefix_map, input_columns)
        # Call the udf and return the output dataframe and columns for this dataframe.
        # We return the columns explicitly since snowflake will auto capitalize them which causes some issues.
        return self._call_udf(ondemand_udf, input_columns, input_df, is_top_level_transformation)

    def _get_dict_keys_from_udf_output(self, output_df: "snowflake.snowpark.DataFrame"):
        from snowflake.snowpark.functions import object_keys

        # Compute the keys from the udf output dict, since we do not know the output schema of the nested transformation.
        # When converting snowpark -> pandas there's some weird formatting we strip out.
        udf_column_names_df = output_df.withColumn(
            "COLUMN_NAMES", object_keys(UDF_OUTPUT_COLUMN_NAME).cast("ARRAY")
        ).limit(1)
        udf_column_names_str = str(udf_column_names_df.to_pandas()["COLUMN_NAMES"].iloc[0])[1:-1]
        udf_column_names_list = []
        for i in udf_column_names_str.split(","):
            value = i.strip()
            udf_column_names_list.append(value.strip('"'))
        return udf_column_names_list

    def _call_udf(
        self,
        ondemand_udf: "snowflake.snowpark.udf.UserDefinedFunction",
        input_columns: List[str],
        input_df: "snowflake.snowpark.DataFrame",
        is_top_level_transformation: bool,
    ) -> ("snowflake.snowpark.DataFrame", list):
        from snowflake.snowpark.functions import array_construct
        from snowflake.snowpark.functions import col

        # Call the udf and append the result to the input dataframe.
        try:
            output_df = input_df.withColumn(UDF_OUTPUT_COLUMN_NAME, ondemand_udf(array_construct(*input_columns)))
        except Exception as e:
            raise UDF_ERROR(e)

        # Rename the feature columns to match the output schema.
        # Only do this for the top level transformation in the pipeline.
        if is_top_level_transformation:
            for column in self._output_schema.keys():
                output_df = output_df.withColumn(
                    self._namespace + SEPERATOR + column if self._append_prefix else column,
                    col(UDF_OUTPUT_COLUMN_NAME)[column].cast(SPARK_TO_SNOWFLAKE_TYPES[self._output_schema[column]]),
                )
            columns_to_select = list(input_df.columns) + [
                self._namespace + SEPERATOR + column if self._append_prefix else column
                for column in self._output_schema.keys()
            ]
            return output_df.select(*columns_to_select), columns_to_select

        udf_column_names_list = self._get_dict_keys_from_udf_output(output_df)
        # Rename the output columns with the transformation prefix since this df will be the input to another transformation.
        feature_columns = []
        for feature in udf_column_names_list:
            feature_name = TRANSFORMATION_COLUMN_PREFIX + feature
            feature_columns.append(feature_name)
            output_df = output_df.withColumn(feature_name, col(UDF_OUTPUT_COLUMN_NAME)[feature])
        columns_to_select = input_df.columns + feature_columns
        return output_df.select(*columns_to_select), columns_to_select

    def _generate_on_demand_udf(
        self,
        transformation_node: TransformationNode,
        input_map: Dict[str, List[str]],
        prefix_map: Dict[str, str],
        input_columns: List[str],
    ) -> "snowflake.snowpark.udf.UserDefinedFunction":
        """Returns the name of the registered udf"""
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = transformation.user_function

        use_pandas = transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS

        def _consume_prefix(s: str, prefix: str) -> str:
            assert s.startswith(prefix), f"{prefix} is not a prefix of {s}"
            return s[len(prefix) :]

        if use_pandas:

            def udf(params: List) -> Dict:
                # This is a requirement from snowflake to track udf
                sys._xoptions["snowflake_partner_attribution"].append("tecton-ai")
                # For udfs that have a data source or feature view as an input, use kwargs.
                # For udfs that have a transformation result as an input, use args.
                kwargs = {}
                args = None
                all_inputs_df = pandas.DataFrame([params], columns=input_columns)
                for input_name, columns in input_map.items():
                    if input_name == TRANSFORMATION_INPUT_PARAMETER:
                        assert args is None, "Only one transformation can be the input to another transformation"
                        df = all_inputs_df[columns]
                        df.columns = df.columns.str[len(prefix_map[input_name]) :]
                        args = df
                    else:
                        df = all_inputs_df[columns]
                        df.columns = df.columns.str[len(prefix_map[input_name]) :]
                        kwargs[input_name] = df

                # Pandas will complain if you use args != None due to how they implement the truth value of a dataframe.
                if not isinstance(args, type(None)):
                    result = user_function(args)
                else:
                    result = user_function(**kwargs)

                result_df = result.astype("object")
                return {col: result_df[col][0] for col in result_df.columns}

        else:

            def udf(params: List) -> Dict:
                sys._xoptions["snowflake_partner_attribution"].append("tecton-ai")
                # For udfs that have a data source or feature view as an input, use kwargs.
                # For udfs that have a transformation result as an input, use args.
                kwargs = {}
                args = None
                all_inputs = dict(zip(input_columns, params))
                for input_name, columns in input_map.items():
                    if input_name == TRANSFORMATION_INPUT_PARAMETER:
                        assert args is None, "Only one transformation can be the input to another transformation"
                        args = {
                            _consume_prefix(column, prefix_map[input_name]): all_inputs[column] for column in columns
                        }
                    else:
                        kwargs[input_name] = {
                            _consume_prefix(column, prefix_map[input_name]): all_inputs[column] for column in columns
                        }
                if args is not None:
                    return user_function(args)

                return user_function(**kwargs)

        if use_pandas:
            # Make sure to update the pandas version in python/requirements.in as well to keep it in sync.
            self._session.add_packages("pandas==1.3.5")
        return self._session.udf.register(udf, replace=True)



import json
import sys

import click

from tecton.cli import printer
from tecton.cli.cli_utils import pprint_dict
from tecton.cli.command import TectonGroup
from tecton.identities import api_keys
from tecton_core.id_helper import IdHelper


@click.command("api-key", cls=TectonGroup)
def api_key():
    """Introspect API key. To create, delete, or list API keys see `tecton service-account` commands"""


def introspect(api_key):
    response = api_keys.introspect(api_key)
    if not response:
        printer.safe_print(
            "API key cannot be found. Ensure you have the correct API Key. The key's secret value is different from the key's ID.",
            file=sys.stderr,
        )
        sys.exit(1)
    return {
        "API Key ID": IdHelper.to_string(response.id),
        "Description": response.description,
        "Created by": response.created_by,
        "Active": response.active,
    }


@api_key.command("introspect")
@click.argument("api-key", required=True)
@click.option(
    "--json-output",
    is_flag=True,
    default=False,
    help="Whether the output is displayed in machine readable json format. Defaults to false.",
)
def introspect_command(api_key, json_output):
    """Introspect an API Key"""
    api_key_details = introspect(api_key)
    if json_output:
        for key in api_key_details.copy():
            snake_case = key.replace(" ", "_").lower()
            api_key_details[snake_case] = api_key_details.pop(key)
        printer.safe_print(f"{json.dumps(api_key_details)}")
    else:
        pprint_dict(api_key_details, colwidth=16)
